# Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import sqlite3
import os.path

from db_generic_re_grt import GenericReverseEngineering

from wb import DefineModule
from workbench.utils import server_version_str2tuple
from workbench.exceptions import NotConnectedError

import grt

ModuleInfo = DefineModule(name= "DbSQLiteRE", author= "Oracle Corp.", version="1.0")

class SQLiteReverseEngineering(GenericReverseEngineering):
    @classmethod
    def getTargetDBMSName(cls):
        return 'SQLite'

    @classmethod
    def serverVersion(cls, connection):
        return cls._connections[connection.__id__]["version"]

    #########  Connection related functions #########
    @classmethod
    def connect(cls, connection, password):
        '''Establishes a connection to the server and stores the connection object in the connections pool.

        It first looks for a connection with the given connection parameters in the connections pool to
        reuse existent connections. If such connection is found it queries the server to ensure that the
        connection is alive and reestablishes it if is dead. If no suitable connection is found in the
        connections pool, a new one is created and stored in the pool.

        Parameters:
        ===========
            connection:  an object of the class db_mgmt_Connection storing the parameters
                         for the connection.
            password:    a string with the password to use for the connection (ignored for SQLite).
        '''
        con = None
        try:
            con = cls.get_connection(connection)
            try:
                if not con.cursor().execute('SELECT 1'):
                    raise Exception('connection error')
            except Exception, exc:
                grt.send_info('Connection to %s apparently lost, reconnecting...' % connection.hostIdentifier)
                raise NotConnectedError('Connection error')
        except NotConnectedError, exc:
            grt.send_info('Connecting to %s...' % connection.hostIdentifier)
            con = sqlite3.connect(connection.parameterValues['dbfile'])
            if not con:
                grt.send_error('Connection failed', str(exc))
                raise
            connection.parameterValues['wbcopytables_connection_string'] = "'" + connection.parameterValues['dbfile'] + "'"
            grt.send_info('Connected')
            cls._connections[connection.__id__] = {'connection': con}
        if con:
            ver = cls.execute_query(connection, "SELECT sqlite_version()").fetchone()[0]
            grt.log_info('SQLite RE', 'Connected to %s, %s' % (connection.name, ver))
            ver_parts = server_version_str2tuple(ver) + (0, 0, 0, 0)
            version = grt.classes.GrtVersion()
            version.majorNumber, version.minorNumber, version.releaseNumber, version.buildNumber = ver_parts[:4]
            cls._connections[connection.__id__]['version'] = version
        return 1


    @classmethod
    def getCatalogNames(cls, connection):
        """Returns a list of the available catalogs.
        """
        return ['def']

    @classmethod
    def getSchemaNames(cls, connection, catalog_name):
        """Returns a list of schemata for the given connection object."""

        return [os.path.splitext(os.path.basename(connection.parameterValues['dbfile']))[0]]

    @classmethod
    def getTableNames(cls, connection, catalog_name, schema_name):
        query = """SELECT name
FROM sqlite_master
WHERE type='table' AND NOT name LIKE 'sqlite_%'"""
        return [row[0] for row in cls.execute_query(connection, query)]

    @classmethod
    def getViewNames(cls, connection, catalog_name, schema_name):
        return []

    @classmethod
    def getProcedureNames(cls, connection, catalog_name, schema_name):
        return []


    #########  Reverse Engineering functions #########


    @classmethod
    def reverseEngineer(cls, connection, catalog_name, schemata_list, context):
        from grt.modules import MysqlSqlFacade
        grt.send_progress(0, "Reverse engineering catalog information")
        cls.check_interruption()
        catalog = cls.reverseEngineerCatalog(connection, catalog_name)

        # calculate total workload 1st
        grt.send_progress(0.1, 'Preparing...')

        get_tables = context.get("reverseEngineerTables", True)

        # 10% of the progress is for preparation
        total = 1e-10  # total should not be zero to avoid DivisionByZero exceptions
        total += len(cls.getTableNames(connection, catalog_name, '')) if get_tables else 0

        grt.send_progress(0.1, "Gathered stats")

        # Now the second pass for reverse engineering tables:
        if get_tables:
            idx = 0
            for object_type, name, tbl_name, _, sql in cls.execute_query(connection, "SELECT * FROM sqlite_master"):
                if type in ('view', 'trigger') or not sql or tbl_name.startswith('sqlite_'):
                    continue

                sql = sql.replace('[', '').replace(']', '')

                grt.log_debug('SQLiteReverseEngineering', 'Processing this sql:\n%s;' % sql)

                MysqlSqlFacade.parseSqlScriptString(catalog, sql)

                cls.check_interruption()
                grt.send_progress(0.1 + idx / total, 'Object %s reverse engineered!' % name)
                idx += 1

        grt.send_progress(1.0, 'Reverse engineering completed!')
        return catalog


    @classmethod
    def reverseEngineerCatalog(cls, connection, catalog_name):
        catalog = grt.classes.db_mysql_Catalog()
        catalog.name = catalog_name
        catalog.simpleDatatypes.extend(cls._rdbms.simpleDatatypes)
        
        schemata_names = cls.getSchemaNames(connection, catalog_name) or ['']
        catalog.schemata.remove_all()
        for schema_name in schemata_names:
            schema = grt.classes.db_mysql_Schema()
            schema.name = schema_name
            schema.owner = catalog
            if hasattr(cls, 'reverseEngineerSequences'):
                cls.reverseEngineerSequences(connection, schema)
            catalog.schemata.append(schema)

        return catalog


    @classmethod
    def getCommentForTable(cls, connection, table):
        pass



    @classmethod
    def reverseEngineerUserDatatypes(cls, connection, catalog):
        pass



###########################################################################################

@ModuleInfo.export(grt.classes.db_mgmt_Rdbms)
def initializeDBMSInfo():
    return SQLiteReverseEngineering.initializeDBMSInfo('sqlite_rdbms_info.xml')

@ModuleInfo.export((grt.LIST, grt.STRING))
def getDataSourceNames():
    return SQLiteReverseEngineering.getDataSourceNames()


@ModuleInfo.export(grt.STRING, grt.STRING)
def quoteIdentifier(name):
    return SQLiteReverseEngineering.quoteIdentifier(name)


@ModuleInfo.export(grt.STRING, grt.classes.GrtNamedObject)
def fullyQualifiedObjectName(obj):
    return SQLiteReverseEngineering.fullyQualifiedObjectName(obj)


@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.STRING)
def connect(connection, password):
    return SQLiteReverseEngineering.connect(connection, password)


@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection)
def disconnect(connection):
    return SQLiteReverseEngineering.disconnect(connection)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection)
def isConnected(connection):
    return SQLiteReverseEngineering.isConnected(connection)

@ModuleInfo.export(grt.STRING)
def getTargetDBMSName():
    return SQLiteReverseEngineering.getTargetDBMSName()

@ModuleInfo.export(grt.LIST)
def getSupportedObjectTypes():
    return SQLiteReverseEngineering.getSupportedObjectTypes()

@ModuleInfo.export(grt.classes.GrtVersion, grt.classes.db_mgmt_Connection)
def getServerVersion(connection):
    return SQLiteReverseEngineering.getServerVersion(connection)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection)
def getCatalogNames(connection):
    return SQLiteReverseEngineering.getCatalogNames(connection)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING)
def getSchemaNames(connection, catalog_name):
    return SQLiteReverseEngineering.getSchemaNames(connection, catalog_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getTableNames(connection, catalog_name, schema_name):
    return SQLiteReverseEngineering.getTableNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getViewNames(connection, catalog_name, schema_name):
    return SQLiteReverseEngineering.getViewNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getTriggerNames(connection, catalog_name, schema_name):
    return SQLiteReverseEngineering.getTriggerNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getProcedureNames(connection, catalog_name, schema_name):
    return SQLiteReverseEngineering.getProcedureNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getFunctionNames(connection, catalog_name, schema_name):
    return SQLiteReverseEngineering.getFunctionNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.classes.db_Catalog, grt.classes.db_mgmt_Connection, grt.STRING, (grt.LIST, grt.STRING), grt.DICT)
def reverseEngineer(connection, catalog_name, schemata_list, context):
    return SQLiteReverseEngineering.reverseEngineer(connection, catalog_name, schemata_list, context)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Catalog)
def reverseEngineerUserDatatypes(connection, catalog):
    return SQLiteReverseEngineering.reverseEngineerUserDatatypes(connection, catalog)

@ModuleInfo.export(grt.classes.db_Catalog, grt.classes.db_mgmt_Connection, grt.STRING)
def reverseEngineerCatalog(connection, catalog_name):
    return SQLiteReverseEngineering.reverseEngineerCatalog(connection, catalog_name)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerTables(connection, schema):
    return SQLiteReverseEngineering.reverseEngineerTables(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerViews(connection, schema):
    return SQLiteReverseEngineering.reverseEngineerViews(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerProcedures(connection, schema):
    return SQLiteReverseEngineering.reverseEngineerProcedures(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerFunctions(connection, schema):
    return SQLiteReverseEngineering.reverseEngineerFunctions(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerTriggers(connection, schema):
    return SQLiteReverseEngineering.reverseEngineerTriggers(connection, schema)



#!/usr/bin/python

import subprocess
import os
import sys
import shutil
import xml.dom.minidom
import xml.dom
import uuid
import stat

WORKING_DIR = os.getcwd()

os.chdir(os.getcwd() + "/../..")

banned_files = ('README.md', 'LICENSE', '.git', 'COPYING', 'NEWS.md')
files_to_keep = ('sys_reports.js')

PROJECT_DIR = os.getcwd() + "/"
SCRIPT_DEPLOYMENT_DIR = PROJECT_DIR + "res/scripts/sys/"
MSI_FRAGMENT_FILE = PROJECT_DIR + "build/msi/source/mysql_workbench_fragment.xml"
CMAKE_FILE = PROJECT_DIR + "res/scripts/CMakeLists.txt"

clone_repo = False

def print_help():
    print 'Usage: python update_wbsys.py [Options]'
    print 'Options:'
    print '  --clone        Force to clone the repository'
    print '  --help         Prints this help'

if len(sys.argv) == 1:
    clone_repo = False
elif len(sys.argv) > 1:
    if sys.argv[1] == '--clone':
        clone_repo = True
    elif sys.argv[1] == '--help':
        print_help()
        exit(0)
    else:
        print 'Unknown option: ' + sys.argv[1]
        print_help()
        exit(0)
else:
    print_help()
    exit(0)


print "Working dir  : " + WORKING_DIR
print "Project dir  : " + PROJECT_DIR
print "MSI Frag file: " + MSI_FRAGMENT_FILE
print "CMake file   : " + CMAKE_FILE
print "Clone repo   : " + str(clone_repo)

# This class searches a MSI XML fragment file for the wbsys 
# files, so that they get installed along with WB. The fragment
# file is updated with the processed wbsys files
class XMLFileManager:
  
    # The constructor reads the fragment file and finds the right
    # component element that will contain the wbsys files
        
    def __init__(self):
        self.directories = dict()
        self.files = dict()
        self.fileCount = 0
        
        self.xml_document = xml.dom.minidom.parse(MSI_FRAGMENT_FILE)
    
        self.root = self.xml_document.documentElement
        directory_elements = self.root.getElementsByTagName("Directory")
  
        for directory_element in directory_elements:
    
            if directory_element.getAttribute("Id") == "sys_dir":
                component = directory_element.getElementsByTagName("Component").item(0)
                
                indent = directory_element.parentNode.firstChild

                # Reuse the base node, so that we have the same ID
                while component.firstChild != None:
                    component.removeChild(component.firstChild)
                    
                while directory_element.firstChild != None:
                    directory_element.removeChild(directory_element.firstChild)
                
                directory_element.appendChild(indent.cloneNode(False))
                
                self.appendNodeChild(component, directory_element)
                
                self.directories['sys'] = directory_element
                
                break

    
    def appendNodeChild(self, child, parent):
      
      parentIndent = parent.lastChild
      
      childIndent = parentIndent.cloneNode(False)
      childIndent.nodeValue = childIndent.nodeValue + '  '
      
      parent.insertBefore(childIndent.cloneNode(False), parent.lastChild)
      parent.insertBefore(child, parent.lastChild)

      ## Don't create child indentation
      if child.tagName == 'File':
        return

      if child.hasChildNodes() == False:
        child.appendChild(childIndent.cloneNode(False))
      elif child.lastChild.nodeType != child.TEXT_NODE:
        child.appendChild(childIndent.cloneNode(False))
      else:
        print 'Node already as indentation'
        
        
    def add_directory(self, directory):
        
        parent_dir = os.path.dirname(directory)
        current_dir = os.path.basename(directory)
        
        #parent_node = self.directories[parent_dir]
        entryId = directory.replace('/', '_')
        
        newEntry = self.xml_document.createElement("Directory")
        newEntry.setAttribute("Id", entryId + '_dir')
        newEntry.setAttribute("Name", current_dir)

        self.appendNodeChild(newEntry, self.directories[parent_dir])        
        self.directories[directory] = newEntry
      
      
    def add_file(self, directory, filename):
        self.fileCount = self.fileCount + 1
        
        newFile = self.xml_document.createElement("File")
        
        newFile.setAttribute('Id', 'wbsys' + '%03d' % self.fileCount)
        newFile.setAttribute('Name', filename)
        
        component = self.directories[directory].getElementsByTagName("Component").item(0)
        
        if component == None:
            component = self.xml_document.createElement("Component")
            component.setAttribute('DiskId', '1')
            component.setAttribute('Guid', str(uuid.uuid1()).upper())
            component.setAttribute('Id', directory.replace('/', '_'))
            component.setAttribute('Location', 'either')
            self.appendNodeChild(component, self.directories[directory])
        
        self.appendNodeChild(newFile, component)
    
    # Overwrites the original XML file
    def write_file(self):

        self.directories['sys'].normalize() 
        for node in self.directories:
          self.directories[node] = self.directories[node].normalize()

        text = self.xml_document.toxml()

        f = open(MSI_FRAGMENT_FILE, mode='w')
        f.write(text.replace("\r\n", "\n"))
        f.close()
           



# This class deals with a CMakeLists.txt file in order
# to setup the proper wbsys files to install.
class CMakeFileManager:
    file_lines_before = None
    file_lines = None
    file_lines_after = None
    state = 0
    
    # The constructor read the file into line arrays, so
    # They can be writen again properly. The automates files
    # are always discarted, and new ones are set during the
    # process.
    def __init__(self):
        with open(CMAKE_FILE) as f:
            self.file_lines_before = []
            self.file_lines_after = []
            self.file_lines = dict()
            
            for line in f:
                if self.state == 0:
                    self.file_lines_before.append(line)
                    if line.startswith("# SYS_BEGIN"):
                        self.state = 1
                elif self.state == 1:
                    if line.startswith("# SYS_END"):
                        self.file_lines_after.append(line)
                        self.state = 2
                else:
                    self.file_lines_after.append(line)

    def add_file(self, filename):
        if not os.path.dirname(filename) in self.file_lines:
            self.file_lines[os.path.dirname(filename)] = []
            
        self.file_lines[os.path.dirname(filename)].append(filename)
        
    def write_file(self):
        # Overwrite the original file
        f = open(CMAKE_FILE, mode='w')
        f.writelines(self.file_lines_before)
        
        for cmake_dir in self.file_lines:
            file_tag = cmake_dir.replace('/', '_').upper() + '_FILES'
            f.write('\nset(' + file_tag + '\n')
        
            for file in self.file_lines[cmake_dir]:
                f.write('    ' + file + '\n')
            
            f.write(')\n\n')
            f.write('install(FILES ${%s} DESTINATION ${WB_PACKAGE_SHARED_DIR}/%s)\n' % (file_tag, cmake_dir))
        
        f.writelines(self.file_lines_after)
        f.close()
        

def remove_repository_files(dest):
    entries = os.listdir(dest)

    current_path = os.getcwd()
    os.chdir(dest)

    for entry in entries:
        if entry in files_to_keep:
            continue
                
        if os.path.isfile(entry):
            os.remove(entry)
        else:
            shutil.rmtree(entry)
    
    os.chdir(current_path)
    
    
def recursive_copy(src, dest):
    if os.path.isdir(src):
      
        # Create dir if it doesn't exist in the destination dir
        if not os.path.isdir(dest):
            os.makedirs(dest)
            
        files = os.listdir(src)

        for file in files:
            recursive_copy(os.path.join(src, file), os.path.join(dest, file))
    else:
        shutil.copyfile(src, dest)
        subprocess.Popen(["sed", "-i", "s/\r//g", dest])
        st = os.stat(dest)
        os.chmod(dest, st.st_mode | stat.S_IEXEC)
        
        
def iterate_path(path):
  
  os.chdir(os.path.basename(path))
  current_path = os.getcwd()

  entries = sorted(os.listdir(os.curdir))

  # This would print all the files and directories
  for entry in entries:
    
      full_entry = os.path.join(path, entry)
    
      if entry in banned_files:
        continue
    
      if os.path.isdir(entry):          
          xml_manager.add_directory(full_entry)
          
          iterate_path(full_entry)
          
          os.chdir(current_path)
      else: 
          xml_manager.add_file(path, entry)
          cmake_manager.add_file(full_entry)
          #print entry          
  
  
  
  
  

# TODO: Remove this when done
os.chdir(WORKING_DIR)
print "Preparing...please be patient..."

if clone_repo == True:
    print "  - Checking git proxy configuration..."
    proc = subprocess.Popen(["git", "config", "--global", "http.proxy"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    out, err = proc.communicate()

    if not out.strip():
        print "    - WARNING: No proxy configured. Try using git config --global http.proxy <proxy>"
    else:
        print "    - Using proxy: " + out.strip()


    print "  - Cloning remote repository..."  

    os.chdir(WORKING_DIR)
    if (os.path.isdir("mysql-sys")):
        shutil.rmtree("mysql-sys")

    proc = subprocess.Popen(["git", "clone", "https://github.com/MarkLeith/mysql-sys"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    out, err = proc.communicate()

    if proc.returncode > 0:
        print "    - ERROR[" + str(proc.returncode) + "] - There was an error on 'git clone'"
        print err
        sys.exit(1)





print "Processing files..."

xml_manager = XMLFileManager()
cmake_manager = CMakeFileManager()

if clone_repo == True:
    if os.path.isdir('sys') == False:
        os.makedirs('sys')
    remove_repository_files('sys')
    recursive_copy('mysql-sys', 'sys')

if os.path.isdir('sys') == False:
    print 'Error: The sys directory does not exist. Please run this script using the --clone parameter.'
    exit(1)

if os.path.isdir('sys/.git'):
    shutil.rmtree(os.path.join('sys', '.git'))

iterate_path('sys')

xml_manager.write_file()
cmake_manager.write_file()





# Copyright (c) 2015 Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import os


class ImportScript:
    def __init__(self):
        self.error_log_name = 'import_errors.log'
        
    @staticmethod
    def create(target_os):
        if target_os == 'windows':
            return ImportScriptWindows()
        elif target_os == 'linux':
            return ImportScriptLinux()
        elif target_os == 'darwin':
            return ImportScriptDarwin()
        else:
            return None

    def generate_import_script(self, connection_args, path_to_file, schema_name): pass

    def get_script_ext(self): pass

    def get_import_cmd(self, table, path_to_file): pass



class ImportScriptWindows(ImportScript):
    def get_import_cmd(self, table, path_to_file):
        return "LOAD DATA INFILE '%s_#####_import/%s' INTO TABLE %s FIELDS TERMINATED BY ',' ENCLOSED BY '';" % (table['target_schema'], path_to_file, table['target_table'])

    def get_script_ext(self): 
        return 'cmd'

    def generate_import_script(self, connection_args, path_to_file, schema_name):
        output = ['@ECHO OFF']

        output.append('echo Started load data. Please wait.')
        
        output.append('SET MYPATH=%%~dp0')
        
        output.append('IF EXIST %%%%MYPATH%%%%%s del /F %%%%MYPATH%%%%%s' % (self.error_log_name, self.error_log_name))
        output.append("SET command=mysql.exe -h127.0.0.1 -P%s -u%s -p -s -N information_schema -e \"SELECT Variable_Value FROM GLOBAL_VARIABLES WHERE Variable_Name = 'datadir'\" 2^>^> %%%%MYPATH%%%%%s" % (connection_args['target_port'], connection_args['target_user'], self.error_log_name))

        output.append('FOR /F "tokens=* USEBACKQ" %%%%F IN ^(^`%%command%%^`^) DO ^(')
        output.append('    SET DADADIR=%%%%F')
        output.append('^)')

        output.append('if %%ERRORLEVEL%% GEQ 1 ^(')
        output.append('    echo Script has failed. See the log file for details.')
        output.append('    exit /b 1')
        output.append('^)')

        output.append('pushd %%DADADIR%%')

        output.append('mkdir %s_#####_import' % schema_name)
        
        output.append('xcopy %%%%MYPATH%%%%*.csv %s_#####_import\* 2^>^> %%%%MYPATH%%%%%s' % (schema_name, self.error_log_name))
        output.append('if %%ERRORLEVEL%% GEQ 1 ^(')
        output.append('    echo Script has failed. See the log file for details.')
        output.append('    exit /b 1')
        output.append('^)')
        
        output.append('xcopy %%%%MYPATH%%%%*.sql %s_#####_import\* 2^>^> %%%%MYPATH%%%%%s' % (schema_name, self.error_log_name))
        output.append('if %%ERRORLEVEL%% GEQ 1 ^(')
        output.append('    echo Script has failed. See the log file for details.')
        output.append('    exit /b 1')
        output.append('^)')
        
        
        output.append('mysql.exe -h127.0.0.1 -P%s -u%s -p ^< %s_#####_import\%s 2^>^> %%%%MYPATH%%%%%s' % (connection_args['target_port'], connection_args['target_user'], schema_name, path_to_file, self.error_log_name))
        output.append('if %%ERRORLEVEL%% GEQ 1 ^(')
        output.append('    echo Script has failed. See the log file for details.')
        output.append('    exit /b 1')
        output.append('^)')
        
        output.append('rmdir %s_#####_import /s /q' % schema_name)
        output.append('echo Finished load data')
        output.append('popd')
        output.append('pause')
        return output



class ImportScriptLinux(ImportScript):
    def get_import_cmd(self, table, path_to_file):
        return "LOAD DATA INFILE '%s_#####_import/%s' INTO TABLE %s FIELDS TERMINATED BY ',' ENCLOSED BY '';" % (table['target_schema'], path_to_file, table['target_table'])

    def get_script_ext(self): 
        return 'sh'

    def generate_import_script(self, connection_args, path_to_file, schema_name):
        output = ['#!/bin/bash']
        output.append('MYPATH=\`pwd\`')
        
        output.append('if [ -f \$MYPATH/%s ] ; then' % self.error_log_name)
        output.append('    rm \$MYPATH/%s' % self.error_log_name) 
        output.append('fi')
        
        output.append("TARGET_DIR=\`MYSQL_PWD=$arg_target_password mysql -h127.0.0.1 -P%s -u%s -s -N information_schema -e 'SELECT Variable_Value FROM GLOBAL_VARIABLES WHERE Variable_Name = \\\"datadir\\\"'\` 2>> \$MYPATH/%s" % (connection_args['target_port'], connection_args['target_user'], self.error_log_name))
        output.append('if [ \$? -ne 0 ];then')
        output.append('   echo Script has failed. See the log file for details.')
        output.append('   exit 1')
        output.append('fi')
        
        output.append('pushd \$TARGET_DIR')
        
        output.append('mkdir %s_#####_import' % schema_name)
        
        output.append('cp \$MYPATH/*.csv %s_#####_import/ 2>> \$MYPATH/%s' % (schema_name, self.error_log_name))
        output.append('if [ \$? -ne 0 ];then')
        output.append('   echo Script has failed. See the log file for details.')
        output.append('   exit 1')
        output.append('fi')
        
        output.append('cp \$MYPATH/*.sql %s_#####_import/ 2>> \$MYPATH/%s' % (schema_name, self.error_log_name))
        output.append('if [ \$? -ne 0 ];then')
        output.append('   echo Script has failed. See the log file for details.')
        output.append('   exit 1')
        output.append('fi')
        
        output.append('echo Started load data. Please wait.')
        output.append('MYSQL_PWD=$arg_target_password mysql -h127.0.0.1 -P%s -u%s < %s_#####_import/%s 2>> \$MYPATH/%s' % (connection_args['target_port'], connection_args['target_user'], schema_name, path_to_file, self.error_log_name))
        output.append('if [ \$? -ne 0 ];then')
        output.append('   echo Script has failed. See the log file for details.')
        output.append('   exit 1')
        output.append('fi')
        
        output.append('echo Finished load data')
        
        output.append('rm -rf %s_#####_import' % schema_name)
        
        output.append('popd')
        #output.append('read -p Press [Enter] key to continue...')
        return output



class ImportScriptDarwin(ImportScript):
    def get_import_cmd(self, table, path_to_file):
        return "LOAD DATA INFILE '%s_#####_import/%s' INTO TABLE %s FIELDS TERMINATED BY ',' ENCLOSED BY '';" % (table['target_schema'], path_to_file, table['target_table'])

    def get_script_ext(self): 
        return 'sh'

    def generate_import_script(self, connection_args, path_to_file, schema_name):
        output = ['#!/bin/bash']
        output.append('MYPATH=\`pwd\`')
        
        output.append('if [ -f \$MYPATH/%s ] ; then' % self.error_log_name)
        output.append('    rm \$MYPATH/%s' % self.error_log_name) 
        output.append('fi')
        
        output.append("TARGET_DIR=\`MYSQL_PWD=$arg_target_password mysql -h127.0.0.1 -P%s -u%s -s -N information_schema -e 'SELECT Variable_Value FROM GLOBAL_VARIABLES WHERE Variable_Name = \\\"datadir\\\"'\` 2>> \$MYPATH/%s" % (connection_args['target_port'], connection_args['target_user'], self.error_log_name))
        output.append('if [ \$? -ne 0 ];then')
        output.append('   echo Script has failed. See the log file for details.')
        output.append('   exit 1')
        output.append('fi')
        
        output.append('pushd \$TARGET_DIR')
        
        output.append('mkdir %s_#####_import' % schema_name)
        
        output.append('cp \$MYPATH/*.csv %s_#####_import/ 2>> \$MYPATH/%s' % (schema_name, self.error_log_name))
        output.append('if [ \$? -ne 0 ];then')
        output.append('   echo Script has failed. See the log file for details.')
        output.append('   exit 1')
        output.append('fi')
        
        output.append('cp \$MYPATH/*.sql %s_#####_import/ 2>> \$MYPATH/%s' % (schema_name, self.error_log_name))
        output.append('if [ \$? -ne 0 ];then')
        output.append('   echo Script has failed. See the log file for details.')
        output.append('   exit 1')
        output.append('fi')
        
        output.append('echo Started load data. Please wait.')
        output.append('MYSQL_PWD=$arg_target_password mysql -h127.0.0.1 -P%s -u%s < %s_#####_import/%s 2>> \$MYPATH/%s' % (connection_args['target_port'], connection_args['target_user'], schema_name, path_to_file, self.error_log_name))
        output.append('if [ \$? -ne 0 ];then')
        output.append('   echo Script has failed. See the log file for details.')
        output.append('   exit 1')
        output.append('fi')
        
        output.append('echo Finished load data')
        
        output.append('rm -rf %s_#####_import' % schema_name)
        
        output.append('popd')
        #output.append('read -p Press [Enter] key to continue...')
        return output



class SourceRDBMS:
    def __init__(self, source_os):
        self.source_os = source_os

    @staticmethod
    def create(source_rdbms, source_os):
        if source_rdbms == 'mssql':
            return SourceRDBMSMssql(source_os)
        elif source_rdbms == 'mysql':
            return SourceRDBMSMysql(source_os)
        elif source_rdbms == 'postgresql':
            return SourceRDBMSPostgresql(source_os)
        elif source_rdbms == 'sqlanywhere':
            return SourceRDBMSSqlAnywhere(source_os)
        else:
            return None

    def get_copy_table_cmd(self, table):pass



class SourceRDBMSMssql(SourceRDBMS):
    def get_copy_table_cmd(self, table, connection_args):
        return 'bcp "SELECT %(select_expression)s FROM %(unquoted_source_schema)s.%(unquoted_source_table)s" queryout %(source_table)s.csv -c -t, -T -S .\%(source_instance)s -U %(source_user)s -P %%arg_source_passwords' % dict(table.items() + connection_args.items())



class SourceRDBMSMysql(SourceRDBMS):
    def get_copy_table_cmd(self, table, connection_args):
        if self.source_os == 'windows':
            return 'mysqldump.exe --login-path=wb_migration_source -t --tab=. %(source_schema)s %(source_table)s --fields-terminated-by=,' % dict(table.items() + connection_args.items())
        else:
            return 'MYSQL_PWD=$arg_source_password mysqldump -h127.0.0.1 -P%(source_port)s -u%(source_user)s -t  --tab=/tmp %(source_schema)s %(source_table)s --fields-terminated-by=\',\'' % dict(table.items() + connection_args.items())

    def get_cfg_editor_cmd(self, connection_args):
        if self.source_os == 'windows':
            return 'mysql_config_editor.exe set --login-path=wb_migration_source -h127.0.0.1 -P%(source_port)s -u%(source_user)s -p' % connection_args


class SourceRDBMSPostgresql(SourceRDBMS):
    def get_copy_table_cmd(self, table, connection_args):
        return 'psql -U %(source_user)s -d %(source_schema)s -c "COPY %(source_table)s TO stdout DELIMITER \',\';" > %(source_table)s.csv' % dict(table.items() + connection_args.items())


class SourceRDBMSSqlAnywhere(SourceRDBMS):
    def get_copy_table_cmd(self, table, connection_args):
        if self.source_os == 'windows':
            return """dbisql.exe -c "DBN=%(source_schema)s;UID=%(source_user)s;PWD=%arg_source_password%" "SELECT * FROM %(source_table)s; OUTPUT TO '%(source_table)s.csv' FORMAT TEXT DELIMITED BY ',' QUOTE '';""" % dict(table.items() + connection_args.items())
        else:
            return """dbisql -c "DBN=%(source_schema)s;UID=%(source_user)s;PWD=$arg_source_password" "SELECT * FROM %(source_table)s; OUTPUT TO '%(source_table)s.csv' FORMAT TEXT DELIMITED BY ',' QUOTE '';""" % dict(table.items() + connection_args.items())



class DataCopyScript:
    def __init__(self):
        self.error_log_name = 'bulk_copy_errors.log'

    @staticmethod
    def create(source_os):
        if source_os == 'windows':
            return DataCopyScriptWindows()
        elif source_os == 'linux':
            return DataCopyScriptLinux()
        elif source_os == 'darwin':
            return DataCopyScriptDarwin()
        else:
            return None

    def generate(self, tables, script_path, source_rdbms):pass



class DataCopyScriptWindows(DataCopyScript):
    def generate(self, tables, connection_args, script_path, source_rdbms, import_script):
        progress = 0 
        total_progress = (3 + len(tables))
        source_schema = tables[0]['source_schema']
        target_schema = tables[0]['target_schema']
        dir_name = 'dump_%s' % source_schema
        log_file = '%s.log' % dir_name
        import_file_name = 'import_%s.%s' % (source_schema, import_script.get_script_ext())
        import_sql_file_name = 'import_%s.sql' % source_schema

        with open(script_path, 'wb+') as f:
            f.write('@ECHO OFF\r\n')
            f.write('SET MYPATH=%~dp0\r\n')
            f.write('IF EXIST %%MYPATH%%%s del /F %%MYPATH%%%s\r\n' % (self.error_log_name, self.error_log_name))
    
            if isinstance(source_rdbms, SourceRDBMSMysql):
                f.write('mysql_config_editor.exe remove --login-path=wb_migration_source 2>> "%%MYPATH%%%s"\r\n' % self.error_log_name)
                f.write('if %ERRORLEVEL% GEQ 1 (\r\n')
                f.write('    echo Script has failed. See the log file for details.\r\n')
                f.write('    exit /b 1\r\n')
                f.write(')\r\n')

                f.write('%s 2>> "%%MYPATH%%%s"\r\n' % (source_rdbms.get_cfg_editor_cmd(connection_args), self.error_log_name))
                f.write('if %ERRORLEVEL% GEQ 1 (\r\n')
                f.write('    echo Script has failed. See the log file for details.\r\n')
                f.write('    exit /b 1\r\n')
                f.write(')\r\n')
                
                f.write("SET command=mysql.exe -h127.0.0.1 -P%s -u%s -p -s -N information_schema -e \"SELECT Variable_Value FROM GLOBAL_VARIABLES WHERE Variable_Name = 'datadir'\" 2>> \"%%MYPATH%%%s\"\r\n" % (connection_args['source_port'], connection_args['source_user'], self.error_log_name))

                f.write('FOR /F "tokens=* USEBACKQ" %%F IN (`%command%`) DO (\r\n')
                f.write('    SET DADADIR=%%F\r\n')
                f.write(')\r\n')
        
                f.write('if %ERRORLEVEL% GEQ 1 (\r\n')
                f.write('    echo Script has failed. See the log file for details.\r\n')
                f.write('    exit /b 1\r\n')
                f.write(')\r\n')
            else:
                f.write("set arg_source_password=\"<put source password here>\"\r\n")
                f.write('SET DADADIR=%TEMP%\\\r\n')
    
            if not isinstance(import_script, ImportScriptWindows):
                f.write("set arg_target_password=\"<put target password here>\"\r\n")
    
            f.write('pushd %DADADIR%\r\n')
            
            f.write('echo [%d %%%%] Creating directory %s\r\n' % (progress, dir_name))
            f.write('mkdir %s\r\n' % dir_name)
            f.write('pushd %s\r\n' % dir_name)
        
            progress = progress + 1
    
            f.write('copy NUL %s\r\n' % import_sql_file_name)
            f.write('echo %s >> %s\r\n' % ('SET SESSION UNIQUE_CHECKS=0;', import_sql_file_name))
            f.write('echo %s >> %s\r\n' % ('SET SESSION FOREIGN_KEY_CHECKS=0;', import_sql_file_name))
            f.write('echo %s >> %s\r\n' % ('use %s;' % target_schema, import_sql_file_name))
            
            f.write('echo [%d %%%%] Start dumping tables\r\n' % (progress * 100 / total_progress))
         
            for table in tables:
                f.write('%s 2>> "%%MYPATH%%%s"\r\n' % (source_rdbms.get_copy_table_cmd(table, connection_args), self.error_log_name))
                f.write('if %ERRORLEVEL% GEQ 1 (\r\n')
                f.write('    echo Script has failed. See the log file for details.\r\n')
                f.write('    exit /b 1\r\n')
                f.write(')\r\n')
                if isinstance(source_rdbms, SourceRDBMSMysql):
                    f.write('rename %s.txt %s.csv\r\n' % (table['source_table'], table['source_table']))
                    f.write('del %s.sql\r\n' % table['source_table'])
                f.write('echo %s >> %s\r\n' % (import_script.get_import_cmd(table, '%s.csv' % (table['source_table'])), import_sql_file_name))
                progress = progress + 1
                f.write('echo [%d %%%%] Dumped table %s\r\n' % (progress * 100 / total_progress, table['source_table']))
    
            f.write('copy NUL %s\r\n' % import_file_name)
            import_file_lines = import_script.generate_import_script(connection_args, import_sql_file_name, target_schema)
            for line in import_file_lines:
                f.write('(echo %s) >> %s\r\n' % (line, import_file_name))
    
            progress = progress + 1
            f.write('echo [%d %%%%] Generated import script %s\r\n' % (progress * 100 / total_progress, import_file_name))
        
            f.write('popd\r\n')
            f.write('set TEMPDIR=%%DADADIR%%%s\r\n' % dir_name)
            f.write('echo Set fso = CreateObject("Scripting.FileSystemObject") > _zipIt.vbs\r\n')
            f.write('echo InputFolder = fso.GetAbsolutePathName(WScript.Arguments.Item(0)) >> _zipIt.vbs\r\n')
            f.write('echo ZipFile = fso.GetAbsolutePathName(WScript.Arguments.Item(1)) >> _zipIt.vbs\r\n')
            f.write('echo CreateObject("Scripting.FileSystemObject").CreateTextFile(ZipFile, True).Write "PK" ^& Chr(5) ^& Chr(6) ^& String(18, vbNullChar) >> _zipIt.vbs\r\n')
            f.write('echo Set objShell = CreateObject("Shell.Application") >> _zipIt.vbs\r\n')
            f.write('echo Set source = objShell.NameSpace(InputFolder).Items >> _zipIt.vbs\r\n')
            f.write('echo objShell.NameSpace(ZipFile).CopyHere(source) >> _zipIt.vbs\r\n')
            f.write('echo Do Until objShell.NameSpace( ZipFile ).Items.Count ^= objShell.NameSpace( InputFolder ).Items.Count >> _zipIt.vbs\r\n')
            f.write('echo wScript.Sleep 200 >> _zipIt.vbs\r\n')
            f.write('echo Loop >> _zipIt.vbs\r\n')
            
            f.write('CScript  _zipIt.vbs  "%%TEMPDIR%%"  "%%DADADIR%%%s.zip" 2>> "%%MYPATH%%%s"\r\n' % (dir_name, self.error_log_name))
            f.write('if %ERRORLEVEL% GEQ 1 (\r\n')
            f.write('    echo Script has failed. See the log file for details.\r\n')
            f.write('    exit /b 1\r\n')
            f.write(')\r\n')
        
            progress = progress + 1
    
            f.write('echo [%d %%%%] Zipped all files to %s.zip file\r\n' % (progress * 100 / total_progress, dir_name))
    
            f.write('xcopy %s.zip %%MYPATH%% 2>> "%%MYPATH%%%s"\r\n' % (dir_name, self.error_log_name))
            f.write('if %ERRORLEVEL% GEQ 1 (\r\n')
            f.write('    echo Script has failed. See the log file for details.\r\n')
            f.write('    exit /b 1\r\n')
            f.write(')\r\n')
            
            f.write('del %s.zip\r\n' % dir_name)    
            f.write('del _zipIt.vbs\r\n')
            f.write('del /F /Q %s\*.*\r\n' % dir_name)
            f.write('rmdir %s\r\n' % dir_name)
            f.write('popd\r\n')
        
            f.write('echo Now you can copy %%MYPATH%%%s.zip file to the target server and run the import script.\r\n' % dir_name)
            f.write('pause\r\n')



class DataCopyScriptLinux(DataCopyScript):
    def generate(self, tables, connection_args, script_path, source_rdbms, import_script):
        progress = 0 
        total_progress = (3 + len(tables))
        source_schema = tables[0]['source_schema']
        target_schema = tables[0]['target_schema']
        dir_name = 'dump_%s' % source_schema
        log_file = '%s.log' % dir_name
        import_file_name = 'import_%s.%s' % (source_schema, import_script.get_script_ext())
        import_sql_file_name = 'import_%s.sql' % source_schema
    
        with open(script_path, 'w+') as f:
            os.chmod(script_path, 0700)
            f.write('#!/bin/bash\n\n')
            f.write('MYPATH=`pwd`\n')
            
            f.write("arg_source_password=\"<put source password here>\"\n")
            f.write("arg_target_password=\"<put target password here>\"\n")
            f.write("""
    if [ -z "$arg_source_password" ] && [ -z "$arg_target_password" ] ; then
        echo WARNING: Both source and target RDBMSes passwords are empty. You should edit this file to set them.
        exit 1
    fi
    """)

            f.write('if [ -f $MYPATH/%s ] ; then\n' % self.error_log_name)
            f.write('    rm $MYPATH/%s\n' % self.error_log_name) 
            f.write('fi\n')

            for table in tables:
                f.write('if [ -f /tmp/%s.txt ];\n' % table['source_table'])
                f.write('then\n')
                f.write('   rm /tmp/%s.txt 2> /dev/null\n' % table['source_table'])
                f.write('   if [ $? -ne 0 ];then\n')
                f.write('       echo "File /tmp/%s.txt already exists. You should remove file before running this script."\n' % table['source_table'])
                f.write('       exit 1\n')
                f.write('   fi\n')
                f.write('fi\n')
            
            f.write('pushd /tmp\n')
            f.write('echo [%d %%] Creating directory %s\n' % (progress, dir_name))
            f.write('mkdir %s\n' % dir_name)
            f.write('pushd %s\n' % dir_name)
    
            progress = progress + 1
    
            f.write('echo "%s" > %s\n' % ('SET SESSION UNIQUE_CHECKS=0;', import_sql_file_name))
            f.write('echo "%s" >> %s\n' % ('SET SESSION FOREIGN_KEY_CHECKS=0;', import_sql_file_name))
            f.write('echo "%s" >> %s\n' % ('use %s;' % target_schema, import_sql_file_name))
    
    
            f.write('echo [%d %%] Start dumping tables\n' % (progress * 100 / total_progress))
    
            for table in tables:
                f.write('%s 2>> $MYPATH/%s\n' % (source_rdbms.get_copy_table_cmd(table, connection_args), self.error_log_name))
                f.write('if [ $? -ne 0 ];then\n')
                f.write('   echo Script has failed. See the log file for details.\n')
                f.write('   exit 1\n')
                f.write('fi\n')
                
                if isinstance(source_rdbms, SourceRDBMSMysql):
                    f.write('cp /tmp/%s.txt %s.csv\n' % (table['source_table'], table['source_table']))
                    f.write('if [ -f /tmp/%s.sql ];\n' % table['source_table'])
                    f.write('then\n')
                    f.write('   rm /tmp/%s.sql\n' % table['source_table'])
                    f.write('fi\n')
                f.write('echo "%s" >> %s\n' % (import_script.get_import_cmd(table, '%s.csv' % (table['source_table'])), import_sql_file_name))
                progress = progress + 1
                f.write('echo [%d %%] Dumped table %s\n' % (progress * 100 / total_progress, table['source_table']))
    
            f.write('touch %s\n' % import_file_name)
            if isinstance(import_script, ImportScriptDarwin) or isinstance(import_script, ImportScriptLinux):
                f.write('chmod +x %s\n' % import_file_name)
            import_file_lines = import_script.generate_import_script(connection_args, import_sql_file_name, target_schema)
            for line in import_file_lines:
                f.write('echo "%s" >> %s\n' % (line, import_file_name))
    
            progress = progress + 1
            f.write('echo [%d %%] Generated import script %s\n' % (progress * 100 / total_progress, import_file_name))
    
            f.write('popd\n')
            f.write('zip -r %s.zip %s 2>> $MYPATH/%s\n' % (dir_name, dir_name, self.error_log_name))
            f.write('if [ $? -ne 0 ];then\n')
            f.write('   echo Script has failed. See the log file for details.\n')
            f.write('   exit 1\n')
            f.write('fi\n')
    
            progress = progress + 1
            f.write('echo [%d %%] Zipped all files to %s.zip file\n' % (progress * 100 / total_progress, dir_name))
    
            f.write('rm -rf %s\n' % dir_name)
            f.write('cp %s.zip $MYPATH 2>> $MYPATH/%s\n' % (dir_name, self.error_log_name))
            f.write('if [ $? -ne 0 ];then\n')
            f.write('   echo Script has failed. See the log file for details.\n')
            f.write('   exit 1\n')
            f.write('fi\n')
            
            f.write('popd\n')
    
            f.write('echo Now you can copy $MYPATH/%s.zip file to the target server, unzip it and run the import script.\n' % dir_name)
            f.write('read -p "Press [Enter] key to continue..."')



class DataCopyScriptDarwin(DataCopyScript):
    def generate(self, tables, connection_args, script_path, source_rdbms, import_script):
        progress = 0 
        total_progress = (3 + len(tables))
        source_schema = tables[0]['source_schema']
        target_schema = tables[0]['target_schema']
        dir_name = 'dump_%s' % source_schema
        log_file = '%s.log' % dir_name
        import_file_name = 'import_%s.%s' % (source_schema, import_script.get_script_ext())
        import_sql_file_name = 'import_%s.sql' % source_schema
    
        with open(script_path, 'w+') as f:
            os.chmod(script_path, 0700)
            f.write('#!/bin/bash\n\n')
            f.write('MYPATH=`pwd`\n')
            
            f.write("arg_source_password=\"<put source password here>\"\n")
            f.write("arg_target_password=\"<put target password here>\"\n")
            f.write("""
    if [ -z "$arg_source_password" ] && [ -z "$arg_target_password" ] ; then
        echo WARNING: Both source and target RDBMSes passwords are empty. You should edit this file to set them.
        exit 1
    fi
    """)

            f.write('if [ -f $MYPATH/%s ] ; then\n' % self.error_log_name)
            f.write('    rm $MYPATH/%s\n' % self.error_log_name) 
            f.write('fi\n')

            for table in tables:
                f.write('if [ -f /tmp/%s.txt ];\n' % table['source_table'])
                f.write('then\n')
                f.write('   rm /tmp/%s.txt 2> /dev/null\n' % table['source_table'])
                f.write('   if [ $? -ne 0 ];then\n')
                f.write('       echo "File /tmp/%s.txt already exists. You should remove file before running this script."\n' % table['source_table'])
                f.write('       exit 1\n')
                f.write('   fi\n')
                f.write('fi\n')
            
            f.write('pushd /tmp\n')
            f.write('echo [%d %%] Creating directory %s\n' % (progress, dir_name))
            f.write('mkdir %s\n' % dir_name)
            f.write('pushd %s\n' % dir_name)
    
            progress = progress + 1
    
            f.write('echo "%s" > %s\n' % ('SET SESSION UNIQUE_CHECKS=0;', import_sql_file_name))
            f.write('echo "%s" >> %s\n' % ('SET SESSION FOREIGN_KEY_CHECKS=0;', import_sql_file_name))
            f.write('echo "%s" >> %s\n' % ('use %s;' % target_schema, import_sql_file_name))
    
    
            f.write('echo [%d %%] Start dumping tables\n' % (progress * 100 / total_progress))
    
            for table in tables:
                f.write('%s 2>> $MYPATH/%s\n' % (source_rdbms.get_copy_table_cmd(table, connection_args), self.error_log_name))
                f.write('if [ $? -ne 0 ];then\n')
                f.write('   echo Script has failed. See the log file for details.\n')
                f.write('   exit 1\n')
                f.write('fi\n')
                
                if isinstance(source_rdbms, SourceRDBMSMysql):
                    f.write('cp /tmp/%s.txt %s.csv\n' % (table['source_table'], table['source_table']))
                    f.write('if [ -f /tmp/%s.sql ];\n' % table['source_table'])
                    f.write('then\n')
                    f.write('   rm /tmp/%s.sql\n' % table['source_table'])
                    f.write('fi\n')
                f.write('echo "%s" >> %s\n' % (import_script.get_import_cmd(table, '%s.csv' % (table['source_table'])), import_sql_file_name))
                progress = progress + 1
                f.write('echo [%d %%] Dumped table %s\n' % (progress * 100 / total_progress, table['source_table']))
    
            f.write('touch %s\n' % import_file_name)
            if isinstance(import_script, ImportScriptDarwin) or isinstance(import_script, ImportScriptLinux):
                f.write('chmod +x %s\n' % import_file_name)
            import_file_lines = import_script.generate_import_script(connection_args, import_sql_file_name, target_schema)
            for line in import_file_lines:
                f.write('echo "%s" >> %s\n' % (line, import_file_name))
    
            progress = progress + 1
            f.write('echo [%d %%] Generated import script %s\n' % (progress * 100 / total_progress, import_file_name))
    
            f.write('popd\n')
            f.write('zip -r %s.zip %s 2>> $MYPATH/%s\n' % (dir_name, dir_name, self.error_log_name))
            f.write('if [ $? -ne 0 ];then\n')
            f.write('   echo Script has failed. See the log file for details.\n')
            f.write('   exit 1\n')
            f.write('fi\n')
    
            progress = progress + 1
            f.write('echo [%d %%] Zipped all files to %s.zip file\n' % (progress * 100 / total_progress, dir_name))
    
            f.write('rm -rf %s\n' % dir_name)
            f.write('cp %s.zip $MYPATH 2>> $MYPATH/%s\n' % (dir_name, self.error_log_name))
            f.write('if [ $? -ne 0 ];then\n')
            f.write('   echo Script has failed. See the log file for details.\n')
            f.write('   exit 1\n')
            f.write('fi\n')
            
            f.write('popd\n')
    
            f.write('echo Now you can copy $MYPATH/%s.zip file to the target server, unzip it and run the import script.\n' % dir_name)
            f.write('read -p "Press [Enter] key to continue..."')



class DataCopyFactory:
    def __init__(self, source_os, target_os, source_rdbms):
        self.source_rdbms = SourceRDBMS.create(source_rdbms, source_os)
        self.datacopy_script = DataCopyScript.create(source_os)
        self.import_script = ImportScript.create(target_os)

    def _unquoteIdentifier(self, identifier):
        if not '.' in identifier:
            return identifier[1:-1]
        else:
            return identifier[identifier.index('.') + 2:-1]

    def unquoteIdentifiers(self, tables):
        for table in tables:
            table['unquoted_source_schema'] = table['source_schema']
            table['unquoted_source_table'] = table['source_table']
            table['source_schema'] = self._unquoteIdentifier(table['source_schema'])
            table['target_schema'] = self._unquoteIdentifier(table['target_schema'])
            table['source_table'] = self._unquoteIdentifier(table['source_table'])
            table['target_table'] = self._unquoteIdentifier(table['target_table'])

    def generate(self, tables, connection_args, script_path):
        self.unquoteIdentifiers(tables)
        self.datacopy_script.generate(tables, connection_args, script_path, self.source_rdbms, self.import_script)




# Copyright (c) 2012, 2015, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import Queue

from wb import DefineModule
import grt
import mforms
import migration_main

ModuleInfo = DefineModule(name= "Migration", author= "Oracle Corp.", version="1.0")

tab_reference = None

#-------------------------------------------------------------------------------
def handle_view_close(view):
    global tab_reference
    tab_reference = None
    view.cleanup()
    return True

#-------------------------------------------------------------------------------
@ModuleInfo.plugin("wb.migration.open", type="standalone", caption= "Migration Plugin",  pluginMenu= "Migration")
@ModuleInfo.export(grt.INT)
def openMigration():
    global tab_reference
    app = mforms.App.get()
    if tab_reference:
        app.select_view(tab_reference)
        return

    mgview = migration_main.Migration()

    app.dock_view(mgview, "maintab")
    mgview.set_title("Migration")

    tab_reference = mgview
    mgview.on_close(lambda: handle_view_close(mgview)) # TODO: create weak ref

    app.set_status_text("Migration Wizard was started")

    return 1


@ModuleInfo.plugin("wb.migration.showTypeMapEditor", type="standalone", caption= "Type Mapping Editor",  pluginMenu= "Migration")
@ModuleInfo.export(grt.INT)
def showTypeMapEditor():
    import datatype_mapping_editor
    return datatype_mapping_editor.run()
    
    



# Copyright (c) 2012, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import mforms
import grt

from workbench.ui import WizardPage, WizardProgressPage


class MigrationOptionsView(WizardPage):
    def __init__(self, main):
        WizardPage.__init__(self, main, "Migration Options")

        self.main.add_wizard_page(self, "ObjectMigration", "Migration Options")

        label = mforms.newLabel("Select options for the migration of the source schema/schemas to MySQL.")
        self.content.add(label, False, True)

        panel = mforms.newPanel(mforms.TitledBoxPanel)
        panel.set_title("DBMS Specific Options")
        self.content.add(panel, False, True)

        self._db_options_box = mforms.newBox(False)
        panel.add(self._db_options_box)
        self._db_options_box.set_padding(12)
        self._db_options_box.set_spacing(8)
        self._db_options = []


    def page_advanced(self, advancing):
        if advancing:
            for item, name, getter in self._db_options:
                self._db_options_box.remove(item)

            if hasattr(self.main.plan.migrationSource.module_re(), "migrationOptions"):
                options = self.main.plan.migrationSource.module_re().migrationOptions(self.main.plan.state)
  
                self._db_options = self.create_options(self._db_options_box, options)
                if not self._db_options:
                    self._db_options_box.show(False)
                else:
                    self._db_options_box.show(True)
    
        return WizardPage.page_advanced(self, advancing)


    def create_options(self, box, options):
        optlist = []
        for option in options:
            cont = None
            if option.paramType == "boolean":
                opt = mforms.newCheckBox()
                opt.set_active(self.defaultValue == "1")
                box.add(opt, False, True)
                getter = opt.get_string_value
            elif option.paramType == "string":
                hbox = mforms.newBox(True)
                hbox.set_spacing(8)
                hbox.add(mforms.newLabel(option.caption), False, True)
                opt = mforms.newTextEntry()
                opt.set_value(option.defaultValue)
                hbox.add(opt, True, True)
                l = mforms.newLabel(option.description)
                l.set_style(mforms.SmallHelpTextStyle)
                hbox.add(l, False, True)
                box.add(hbox, False, True)
                cont = hbox
                getter = opt.get_string_value
            else:
                grt.send_error("MigrationWizard", "migrationOption() for source has an invalid parameter of type %s (%s)" % (option.paramType, option.name))
                continue
            optlist.append((cont or opt, option.name, getter))
        return optlist


    def go_next(self):
        dic = self.main.plan.state.objectMigrationParams
        for item, name, getter in self._db_options:
            dic[name] = getter()
        WizardPage.go_next(self)



class MigrationProgressView(WizardProgressPage):
    def __init__(self, main):
        WizardProgressPage.__init__(self, main, "Migration", description="""Reverse engineered objects from the source RDBMS will now be automatically 
converted into MySQL compatible objects. Default datatype and default column value
mappings will be used. You will be able to review and edit generated objects and column
definitions in the Manual Editing step.""")
        
        self._autostart = True
        self.add_threaded_task(self.task_migrate, "Migrate Selected Objects")
        self.add_threaded_task(self.task_generate_sql, "Generate SQL CREATE Statements")

        self.main.add_wizard_page(self, "ObjectMigration", "Migration")
 
 
    def task_migrate(self):
        self.main.plan.migrate()
        
        
    def task_generate_sql(self):
        grt.send_progress(0, "Generating SQL...")
        self.main.plan.generateSQL()


    def go_back(self):
        self.reset()
        WizardProgressPage.go_back(self)




# Copyright (c) 2007, 2015, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import threading
from threading import currentThread
import sys
import os
import ntpath
import posixpath

import grt

import mforms

from wb_server_management import wbaOS
from wb_common import OperationCancelledError, InvalidPasswordError, Users

import re

#===============================================================================

UnixVariant = {
    "" : {
        'sudo_command' : "/usr/bin/sudo -S -p EnterPasswordHere ",
    }
}

#===============================================================================
class ServerProfile(object):
    def __init__(self, connection, instance_profile, force_remote_admin_off=False):
        self.disable_remote_admin = force_remote_admin_off
        self.__connection = connection
        self.__settings = instance_profile
        self.__os_specifics = None # one of UnixVariant
        self.expanded_config_file_path = None

    def get_settings_object(self):
        return self.__settings

    def _int_server_info(self, name, default_value=0):
        # always returns an int, a None value will be replaced with default_value
        if not self.__settings:
            return default_value
        try:
            value = self.__settings.serverInfo[name]
        except KeyError:
            value = default_value
        if type(value) is not int:
            try:
                value= int(value)
            except:
                value= default_value
        if value is None:
            return default_value
        return value

    def _int_login_info(self, name, default_value=0):
        # always returns an int, a None value will be replaced with default_value
        try:
            value = self.__settings.loginInfo[name]
        except KeyError:
            value = default_value
        if type(value) is not int:
            try:
                value= int(value)
            except:
                value= default_value
        if value is None:
            return default_value
        return value

    def _str_server_info(self, name, default_value=""):
        # always returns an str or unicode, a None value will be replaced with default_value
        if not self.__settings:
            return default_value
        try:
            value= self.__settings.serverInfo[name]
        except KeyError:
            value = default_value
        if value is None:
            value= default_value
        elif type(value) is not str and type(value) is not unicode:
            value= str(value)
        return value

    def _str_login_info(self, name, default_value=""):
        # always returns an str or unicode, a None value will be replaced with default_value
        try:
            value= self.__settings.loginInfo[name]
        except KeyError:
            value = default_value
        if value is None:
            value= default_value
        elif type(value) is not str and type(value) is not unicode:
            value= str(value)
        return value

    @property
    def name(self):
        return self.__connection.name

    @property
    def db_connection_params(self):
        return self.__connection

    @property
    def host_os(self):
        if hasattr(sys, 'getwindowsversion'):
            return wbaOS.windows
        elif ('inux' in sys.platform):
            return wbaOS.linux
        elif ('arwin' in sys.platform):
            return wbaOS.darwin
        return wbaOS.unknown

    # Must return smth from wbaOS
    @property
    def target_os(self):
        # This call must return either value from profile for remote server
        # If remote server is not set, which means local server, then returned value
        # must be equal to self.host_os
        ret = wbaOS.unknown

        try:
            system = self.__settings.serverInfo['sys.system']
        except:
            system = None
        if system is not None and type(system) is str or type(system) is unicode:
            system = system.strip(" \r\t\n").lower()
            if system == 'windows':
                ret = wbaOS.windows
            elif system in ('linux', 'freebsd', 'opensolaris'):
                ret = wbaOS.linux
            elif ('macos' in system):
                ret = wbaOS.darwin
        else:
            ret = self.host_os

        return ret

    @property
    def host_is_windows(self):
        return self.host_os == wbaOS.windows

    @property
    def target_is_windows(self):
        return self.target_os == wbaOS.windows


    @property
    def remote_admin_enabled(self):
        return (self.uses_ssh or (not self.is_local and self.uses_wmi)) and not self.disable_remote_admin

    @property
    def admin_enabled(self):
        return self.is_local or self.remote_admin_enabled

    @property
    def is_local(self):
        local_addrs = ["localhost", "", "0", "127.0.0.1", "::1"]
        #if self.uses_wmi:
        #    return self.wmi_hostname in local_addrs
        #elif self.uses_ssh:
        #    return self.ssh_hostname in local_addrs
        #else:
        if True:
            if self.db_connection_params:
                params = self.db_connection_params.parameterValues
                if params:
                    if self.db_connection_params.driver:
                        if self.db_connection_params.driver.name == "MysqlNative":
                            return params["hostName"] in local_addrs
                        elif self.db_connection_params.driver.name == "MysqlNativeSocket":
                            return True
                        elif self.db_connection_params.driver.name == "MysqlNativeSSH":
                            return False
                    else:
                        from wb_common import NoDriverInConnection
                        raise NoDriverInConnection, """Workbench has not found a driver for the connection
that is being used by this server instance.
Please edit your connection settings and try again."""
            return False


    @property
    def connect_method(self):
        # choices are ['local', 'ssh', 'wmi', 'none']
        # 'none' is for remote server ip with "No Remote Management" selected
        ret = None
        if self.uses_wmi:
            ret = 'wmi'
        elif self.uses_ssh:
            ret = 'ssh'
        else:
            if self.is_local:
                ret = 'local'
            else:
                ret = 'none'
        return ret

    @property
    def host_name(self):
        host_name = None
        if self.db_connection_params:
            if self.db_connection_params.driver.name == "MysqlNativeSocket":
                host_name = "localhost"
            else:
                host_name = self.db_connection_params.parameterValues["hostName"]
        return host_name
    
    @property
    def port(self):
        port = 3306
        if self.db_connection_params:
            if self.db_connection_params.driver.name != "MysqlNativeSocket":
                port = self.db_connection_params.parameterValues["port"]
        return port

    @property
    def mysql_username(self):
        if self.db_connection_params:
            return self.db_connection_params.parameterValues["userName"]
        return ""

    @property
    def uses_ssh(self):
        return self._int_server_info("remoteAdmin")

    # If WMI admin was selected as mgmt mean (can be for local or remote)
    @property
    def uses_wmi(self):
        return self._int_server_info("windowsAdmin")

    @property
    def ssh_username(self):
        return self._str_login_info("ssh.userName")

    @property
    def ssh_hostname(self):
        return self._str_login_info("ssh.hostName")

    @property
    def ssh_port(self):
        return self._int_login_info("ssh.port", 22)

    @property
    def ssh_usekey(self):
        return self._int_login_info("ssh.useKey")

    @property
    def ssh_key(self):
        return self._str_login_info("ssh.key")
    
    @property
    def ssh_keepalive(self):
        return grt.root.wb.options.options['sshkeepalive'] if grt.root.wb.options.options['sshkeepalive'] is not None else 0

    @property
    def ssh_timeout(self):
        return grt.root.wb.options.options['sshtimeout'] if grt.root.wb.options.options['sshtimeout'] is not None else 10

    @property
    def wmi_hostname(self):
        return self._str_login_info("wmi.hostName")

    @property
    def wmi_username(self):
        return self._str_login_info("wmi.userName")

    @property
    def wmi_service_name(self):
        return self._str_server_info("sys.mysqld.service_name")

    @property
    def start_server_cmd(self):
        ret = None
        if self.uses_wmi:
            ret = "sc start " + self.wmi_service_name
        else:
            ret = self._str_server_info("sys.mysqld.start")
        return ret

    @property
    def stop_server_cmd(self):
        ret = None
        if self.uses_wmi:
            ret = "sc stop " + self.wmi_service_name
        else:
            ret = self._str_server_info("sys.mysqld.stop")
        return ret

    @property
    def use_sudo(self):
        return self._int_server_info("sys.usesudo")

    @property
    def config_file_path(self):
        """ path to configuration file (eg my.cnf)
        may contain path variables that need to be expanded, such as %ProgramPath%
        """
        if self.expanded_config_file_path:
            return self.expanded_config_file_path
        return self._str_server_info('sys.config.path', '').strip(" \r\t\n\b")


    @property
    def config_file_section(self):
        return self._str_server_info('sys.config.section', '').strip(" \r\t\n\b")

    #-------------------------------------------------------
    def server_version(self):
        return self._str_server_info("serverVersion", None)

    def set_server_version(self, version):
        self.__settings.serverInfo["serverVersion"] = version

    server_version = property(server_version, set_server_version)
    #-------------------------------------------------------
    @property
    def server_version_tuple(self):
        if not self.server_version:
            return ()
        version_tokens = self.server_version.split('.')
        ver = []
        for token in version_tokens:
            match = re.match(r'(\d+).*', token)
            if match:
                ver.append(int(match.group(1)))
        return tuple(ver)

    #-------------------------------------------------------
    def log_output(self):
        return self._str_server_info("logOutput")

    def set_log_output(self, output):
        self.__settings.serverInfo['logOutput'] = output

    log_output = property(log_output, set_log_output)

    #-------------------------------------------------------
    def general_log_enabled(self):
        return self._int_server_info("generalLogEnabled")

    def set_general_log_enabled(self, value):
        self.__settings.serverInfo['generalLogEnabled'] = value

    general_log_enabled = property(general_log_enabled, set_general_log_enabled)

    #-------------------------------------------------------
    def slow_log_enabled(self):
        return self._int_server_info("slowLogEnabled")

    def set_slow_log_enabled(self, value):
        self.__settings.serverInfo['slowLogEnabled'] = value

    slow_log_enabled = property(slow_log_enabled, set_slow_log_enabled)

    #-------------------------------------------------------
    def general_log_file_path(self):
        return self._str_server_info("generalLogFilePath")

    def set_general_log_file_path(self, path):
        self.__settings.serverInfo['generalLogFilePath'] = path

    general_log_file_path = property(general_log_file_path, set_general_log_file_path)

    #-------------------------------------------------------
    def slow_log_file_path(self):
        return self._str_server_info("slowLogFilePath")

    def set_slow_log_file_path(self, path):
        self.__settings.serverInfo['slowLogFilePath'] = path

    slow_log_file_path = property(slow_log_file_path, set_slow_log_file_path)

    #-------------------------------------------------------
    def error_log_file_path(self):
        return self._str_server_info("errorLogFilePath")

    def set_error_log_file_path(self, path):
        self.__settings.serverInfo['errorLogFilePath'] = path

    error_log_file_path = property(error_log_file_path, set_error_log_file_path)

    #-------------------------------------------------------
    def datadir(self):
        return self._str_server_info('datadir')

    def set_datadir(self, path):
        self.__settings.serverInfo['datadir'] = path

    datadir = property(datadir, set_datadir)

    #-------------------------------------------------------
    def basedir(self):
        return self._str_server_info('basedir')

    def set_basedir(self, path):
        self.__settings.serverInfo['basedir'] = path

    basedir = property(basedir, set_basedir)

    #-------------------------------------------------------
    def set_os_version_info(self, os_type, os_name, os_variant, os_version):
        self.__settings.serverInfo['detected_os_type'] = os_type
        self.__settings.serverInfo['detected_os_name'] = os_name
        self.__settings.serverInfo['detected_os_variant'] = os_variant
        self.__settings.serverInfo['detected_os_version'] = os_version

    @property
    def detected_os_type(self):
        return self.__settings.serverInfo['detected_os_type']

    @property
    def detected_os_name(self):
        return self.__settings.serverInfo['detected_os_name']

    @property
    def detected_os_variant(self):
        return self.__settings.serverInfo['detected_os_variant']

    @property
    def detected_os_version(self):
        return self.__settings.serverInfo['detected_os_version']

    @property
    def path_module(self):
        return ntpath if self.target_os == wbaOS.windows else posixpath

    def join_target_path(self, path, *comps):
        return self.path_module.join(path, *comps)

    # OS specific things
    @property
    def sudo_prefix(self):
        prefix = UnixVariant['']['sudo_command']

        # If the sudo override is configured, will use it instead
        settings_object = self.get_settings_object()
        serverInfo = settings_object.serverInfo
        
        if serverInfo.has_key('sys.mysqld.sudo_override'):
          sudo_override = serverInfo['sys.mysqld.sudo_override']
          
          if sudo_override.strip():
              prefix = sudo_override

        return prefix


class PasswordHandler(object):
    def __init__(self, server_profile):
        self.server_profile = server_profile
        self.pwd_store = {}

    def get_password_type(self, service_type):
        profile = self.server_profile
        password_type = None
        if service_type == "file":
            if profile.target_is_windows:
                # in windows, auth is handled by UAC which is done externally
                password_type = "UAC"
            else:
                password_type = "sudo"
        elif service_type == "service.startstop":
            if profile.target_is_windows:
                password_type = "wmi"
            else:
                if profile.use_sudo:
                    password_type = "sudo"
                else:
                    password_type = None
        elif service_type == "remoteshell":
            if profile.uses_ssh:
                if profile.ssh_usekey:
                    password_type = "sshkey"
                else:
                    password_type = "ssh"
        elif service_type == "ssh":
            password_type = "ssh"
        elif service_type == "sshkey":
            password_type = "sshkey"
        elif service_type == "local":
            password_type = "local"
        else:
            raise Exception("Unknown password type: %s" % service_type)
        return password_type

    def get_password_parameters(self, service_type):
        # known service types: remoteshell, ssh, sshkey, file, service.startstop
        # these are mapped to
        # known password types: ssh, sshkey, sudo, UAC, wmi
        # which are then mapped to (title, service, account) or "UAC" or None

        profile = self.server_profile

        password_type = self.get_password_type(service_type)

        # sudo pwd may be either user's password, or root password. It depends on sudo setup
        if password_type == "sudo":
            sudo_type = "sudo"
            # if su is to be used instead of sudo, then we need the root password instead of login user
            if profile.sudo_prefix and "sudo" not in profile.sudo_prefix:
                sudo_type = "root"

            if profile.uses_ssh:
                if sudo_type == "sudo":
                    account = profile.ssh_username
                else:
                    account = "root"
                title = "%s Password Required"%sudo_type
                service = "%s@%s"%(sudo_type, profile.ssh_hostname)
            else:
                if sudo_type == "sudo":
                    import pwd
                    # os.getlogin() raises errno 25 if WB is started from ubuntu menu, so never use it
                    #account = os.getlogin()
                    try:
                        account = pwd.getpwuid(os.getuid())[0]
                        if not account:
                            account = "sudo"
                    except:
                        account = "sudo"
                else:
                    account = "root"
                title = "%s Password Required"% sudo_type
                service = "%s@localhost" % sudo_type

            return (title, service, account)

        elif password_type == "local":
            return ("", "Mysql@%s:%s" % (profile.host_name, profile.port), profile.mysql_username)

        elif password_type == "wmi":
            return ("WMI Password Required", "wmi@%s" % profile.wmi_hostname, profile.wmi_username)

        elif password_type == "UAC":
            return "UAC"

        elif password_type == "ssh":
            return ("SSH Login Password Required", "ssh@%s" % profile.ssh_hostname, profile.ssh_username)

        elif password_type == "sshkey":
            return ("SSH Private Key Password Required", "ssh_keyfile@%s" % profile.ssh_key, profile.ssh_username)

        else:
            return None

    # call this if a InvalidPasswordError is caught
    def reset_password_for(self, service_type):
        password_type = self.get_password_type(service_type)
        if self.pwd_store.has_key(password_type):
            # None means the stored password was bad
            self.pwd_store[password_type] = None


    def get_password_for(self, service_type, cached_only=False): # Raises OperationCancelledError if user cancels pwd dlg.
        force_reset = False
        password_type = self.get_password_type(service_type)

        if self.pwd_store.has_key(password_type):
            if self.pwd_store[password_type] is None:
                force_reset = True
            else:
                return self.pwd_store[password_type]

        details = self.get_password_parameters(service_type)
        if not details: # No password needed for this service_type
            return None

        # special case for UAC used in windows
        if details == "UAC":
            return "UAC"

        title, service, account = details

        grt.log_debug2('%s:get_password_for()' % self.__class__.__name__, 'request password for "%s" => "%s", "%s", "%s"\n' % (service_type, title, service, account))

        if cached_only:
            found, password = mforms.Utilities.find_password(service, account)
            if found:
                return password
            return None
        else:
            accepted, password = mforms.Utilities.find_or_ask_for_password(title, service, account, force_reset)
            if accepted:
                if not password:
                    password = ""
                self.pwd_store[password_type] = password.encode("utf8")
                return password
            else:
                raise OperationCancelledError("Password input cancelled")


class ServerControlBase(object):
    def __init__(self, serverProfile, helper, password_delegate):
        self.profile = serverProfile
        self.helper = helper
        self.password_delegate = password_delegate
        self.output_buffer = []
        # default handler will just add the output to a list to be flushed later
        self.output_handler = lambda line, output_buffer= self.output_buffer: output_buffer.append(line)

    def set_output_handler(self, handler): # must be called by consumer of output (ie, the start/stop Tab in WB)
        self.output_handler = handler
        if self.output_buffer:
            for line in self.output_buffer:
                self.output_handler(line)
        self.output_buffer = None

    def info(self, text):
        self.output_handler(text)

    def close(self):
        # do cleanup, free up resources etc
        pass

    def start_async(self, finish_callback, try_without_password): # callback(status) where status can be success, bad_password or an error message
        raise NotImplementedError()

    def stop_async(self, finish_callback, try_without_password):
        raise NotImplementedError()

    def worker_thread(self, action, password, finish_callback):
        try:
            action(password)
            finish_callback("success")
        except InvalidPasswordError, err:
            if password is None:
                finish_callback("need_password")
            else:
                finish_callback("bad_password")
        except Exception, err:
            import traceback
            traceback.print_exc()
            finish_callback(err)

    def get_status(self, verbose=0): # overridden by concrete subclasses, if available
        return None


class ServerControlShell(ServerControlBase):
    def __init__(self, profile, helper, password_delegate):
        """ Can also throw OperationCancelledError"""
        ServerControlBase.__init__(self, profile, helper, password_delegate)

    def get_password(self, try_without_password=False):
        return self.password_delegate.get_password_for("service.startstop", cached_only=try_without_password)

    def start_async(self, finish_callback, try_without_password=False): # callback(status) where status can be success, bad_password or an error message
        if not self.profile.start_server_cmd:
            self.info("Command to start the server is not configured. Please set the command that must be used to start the server in the remote management section of this connections settings.")
            return False
        if self.profile.use_sudo:
            try:
                password = self.get_password(try_without_password)
            except OperationCancelledError:
                return False
        else:
            password = None
        thread = threading.Thread(target=self.worker_thread, args=(self.start, password, finish_callback))
        thread.run()
        return True

    def stop_async(self, finish_callback, try_without_password=False):
        if not self.profile.stop_server_cmd:
            self.info("Command to stop the server is not configured. Please set the command that must be used to stop the server in the remote management section of this connections settings.")
            return False
        if self.profile.use_sudo:
            try:
                password = self.get_password(try_without_password)
            except OperationCancelledError:
                return False
        else:
            password = None
        thread = threading.Thread(target=self.worker_thread, args=(self.stop, password, finish_callback))
        thread.run()
        return True
    
    def start(self, password):
        """ Can also throw InvalidPasswordError"""
        self.info("Starting server...") # TODO: This does not come to info output
        try:
            self.info("Executing '%s'" % self.profile.start_server_cmd)
            as_user = Users.ADMIN if self.profile.use_sudo else Users.CURRENT
            r = self.helper.execute_command(self.profile.start_server_cmd,
                                    as_user=as_user,
                                    user_password=password,
                                    output_handler=lambda s:self.info("Start server: %s"%(s if ((type(s) is str) or (type(s) is unicode)) else "").replace(password or "", "")))
        except InvalidPasswordError:
            # forget password, so that next time it will be requested
            self.password_delegate.reset_password_for("service.startstop")
            raise

        if r:
            return True
        return False


    def stop(self, password):
        """ Can also throw InvalidPasswordError"""
        self.info("Stopping server...")
        try:
            def strip_pwd(output):
                if output is not None and (type(output) is str or type(output) is unicode):
                    if password is not None and password != "":
                        output = output.replace(password, "")
                else:
                    output = ""
                self.info("Stop server: %s"%str(output))

            as_user = Users.ADMIN if self.profile.use_sudo else Users.CURRENT
            self.info("Executing '%s'" % self.profile.stop_server_cmd)
            r = self.helper.execute_command(self.profile.stop_server_cmd,
                                    as_user=as_user,
                                    user_password=password,
                                    output_handler=strip_pwd)
        except InvalidPasswordError:
            # forget password, so that next time it will be requested
            self.password_delegate.reset_password_for("service.startstop")
            raise

        if r:
            return True
        return False


class ServerControlWMI(ServerControlBase):
    def __init__(self, profile, helper, password_delegate):
        ServerControlBase.__init__(self, profile, helper, password_delegate)
        self.wmi = grt.modules.Workbench
        self.wmi_session_ids = {}
        self.shell = None
        if self.profile.is_local:
            user = ""
            server = ""
            password = ""
            self.check_and_fix_profile_for_local_windows(profile)
            self.shell = ServerControlShell(profile, helper, password_delegate)
            # Force usage of Users.ADMIN for local windows
            self.info("Workbench will use cmd shell commands to start/stop this instance")
        else:
            user = self.profile.wmi_username
            server = self.profile.wmi_hostname
            password = self.get_password()
        try:
            sess = self.wmi.wmiOpenSession(server, user, password or "")
            self.wmi_session_ids[currentThread()] = sess
        except Exception, exc:
            import traceback
            traceback.print_exc()
            raise RuntimeError("Could not initialize WMI interface: %s"%exc)
        if self.wmi_session_id_for_current_thread <= 0:
            raise RuntimeError("Could not initialize WMI interface")


    def check_and_fix_profile_for_local_windows(self, profile):
        settings = profile.get_settings_object()
        serverInfo = settings.serverInfo

        serverInfo["sys.usesudo"] = 1 # Force this in any case
        # Check commands
        service = profile.wmi_service_name
        #  return self._str_server_info("sys.mysqld.service_name")
        if service == "":
            service = serverInfo["sys.mysqld.service_name"] = "MySQL"
            self.info("MySQL service was empty. Set to 'MySQL'. Check this in 'Manage Server Instances' from 'Home'.")

        if profile.start_server_cmd == "" or (service not in profile.start_server_cmd):
            serverInfo["sys.mysqld.start"] = "sc start " + service
            print "WMIShell: start command set to '%s'"%profile.start_server_cmd

        if profile.stop_server_cmd == "" or (service not in profile.stop_server_cmd):
            serverInfo["sys.mysqld.stop"] = "sc stop " + service
            print "WMIShell: stop command set to '%s'"%profile.stop_server_cmd


    def get_password(self):
        if self.profile.is_local:
            return ""
        return self.password_delegate.get_password_for("service.startstop")

    # WMI sessions are thread specific. To make things easier, we create one for
    # each thread that uses it
    @property
    def wmi_session_id_for_current_thread(self):
        thr = currentThread()
        s = self.wmi_session_ids.get(thr)
        if s is not None:
            return s
        password = self.get_password()

        if self.profile.is_local:
            sess = self.wmi.wmiOpenSession('', '', '')
        else:
            sess = self.wmi.wmiOpenSession(self.profile.wmi_hostname, self.profile.wmi_username, password or "")

        if sess:
            self.wmi_session_ids[thr] = sess

        return sess

    def close(self):
        for s in self.wmi_session_ids.values():
            self.wmi.wmiCloseSession(s)
        self.wmi_session_ids = {}
        ServerControlBase.close() # not certain if the call is needed.

    def start_async(self, finish_callback, try_without_password): # callback(status) where status can be success, bad_password or an error message
        try:
            # ask subclass to provide password
            password = self.get_password()
        except OperationCancelledError:
            return False
        thread = threading.Thread(target=self.worker_thread, args=(self.start, password, finish_callback))
        thread.run()
        return True

    def stop_async(self, finish_callback, try_without_password):
        try:
            password = self.get_password()
        except OperationCancelledError:
            return False
        thread = threading.Thread(target=self.worker_thread, args=(self.stop, password, finish_callback))
        thread.run()
        return True
    
    def start(self, password):
        self.info("Starting server...");
        if self.shell:
            return self.shell.start(password)
        else:
            # Profile may hold the actual command passed
            service = self.profile.wmi_service_name
            action = "start"
            self.info("Starting service '%s' through WMI..." % service)
            result = self.wmi.wmiServiceControl(self.wmi_session_id_for_current_thread, service, action)
            self.info("Service start result: %s" % result)
            if result.startswith("error"):
                raise RuntimeError("Error stopping service %s through WMI\n%s" % (service, result))
            return not result.startswith("error")


    def stop(self, password):
        if self.shell:
            return self.shell.stop(password)
        else:
            service = self.profile.wmi_service_name
            action = "stop"
            self.info("Stopping service '%s' through WMI..." % service)
            result = self.wmi.wmiServiceControl(self.wmi_session_id_for_current_thread, service, action)
            self.info("Service stop result: %s" % result)
            if result.startswith("error"):
                raise RuntimeError("Error stopping service %s through WMI\n%s" % (service, result))
            return not result.startswith("error")


    def get_status(self, verbose=1):
        "Returned value is one of running, stopping, starting, stopped, unknown"
        service = self.profile.wmi_service_name
        action = "status"
        if verbose > 1:
            self.info("Checking service status of instance %s..." % service)
        result = self.wmi.wmiServiceControl(self.wmi_session_id_for_current_thread, service, action)
        if verbose:
            self.info("Status check of service '%s' returned %s" % (service, result))
        # translate status to something more displayable
        if result == "stop pending":
            result = "stopping"
        elif result == "start pending":
            result = "starting"

        return result



# Copyright (c) 2009, 2013, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import os

import grt
import mforms
from wb import wbinputs
from wb_utils_grt import ModuleInfo


# this is just a function used by the plugin, it's not exported
def printTableLine(fields, filler= " "):
    print "|",
    for text, size in fields:
        print text.ljust(size, filler), "|",
    print


# @wbexport makes this function be exported by the module and also describes the return and
# argument types of the function
# @wbplugin defines the name of the plugin to "wb.catalog.util.dumpColumns", sets the caption to be
# shown in places like the menu, where to take input arguments from and also that it should be included
# in the Catalog submenu in Plugins.
@ModuleInfo.plugin("wb.catalog.util.dumpColumns", caption= "Dump All Table Columns", input= [wbinputs.currentCatalog()], pluginMenu= "Catalog")
@ModuleInfo.export(grt.INT, grt.classes.db_Catalog)
def printAllColumns(catalog):
    lines= []
    schemalen= 0
    tablelen= 0
    columnlen= 0
    typelen= 0

    for schema in catalog.schemata:
        schemalen= max(schemalen, len(schema.name))
        for table in schema.tables:
            tablelen= max(tablelen, len(table.name))
            for column in table.columns:
                columnlen= max(columnlen, len(column.name))
                typelen= max(typelen, len(column.formattedType))
                lines.append((schema.name, table.name, column.name, column.formattedType))

    printTableLine([("-", schemalen), ("-", tablelen), ("-", columnlen), ("-", typelen)], "-")
    printTableLine([("Schema", schemalen), ("Table", tablelen), ("Column", columnlen), ("Type", typelen)])

    printTableLine([("-", schemalen), ("-", tablelen), ("-", columnlen), ("-", typelen)], "-")

    for s,t,c,dt in lines:
        printTableLine([(s, schemalen), (t, tablelen), (c, columnlen), (dt, typelen)])

    printTableLine([("-", schemalen), ("-", tablelen), ("-", columnlen), ("-", typelen)], "-")
    print len(lines), "columns printed"
    
    return 0



@ModuleInfo.plugin("wb.model.print_diagram_pdf", caption= "Export Diagram to PDF...", input= [wbinputs.selectedDiagram()], pluginMenu="Overview")
@ModuleInfo.export(grt.INT, grt.classes.model_Diagram)
def printDiagramToPDF(diagram):
    fc = mforms.FileChooser(mforms.Form.main_form(), mforms.SaveFile)
    fc.set_title("Export to PDF...")
    fc.set_extensions("PDF files (*.pdf)|*.pdf", ".pdf")
    if fc.run_modal():
        path = fc.get_path()
        mforms.App.get().set_status_text("Exporting diagram %s to PDF file..." % diagram.name)
        grt.modules.WbPrinting.printToPDFFile(diagram, path)
        mforms.App.get().set_status_text("Diagram %s exported to %s" % (diagram.name, path))
    return 0



class PrintToFileDialog(mforms.Form):
    def __init__(self, model):
        mforms.Form.__init__(self, mforms.Form.main_form(), mforms.FormNormal)
        self.set_title("Print Model to File")

        self.model = model

        box = mforms.newBox(False)
        box.set_spacing(8)
        box.set_padding(20)

        self.tree = mforms.newTreeNodeView(mforms.TreeFlatList)
        self.tree.add_column(mforms.CheckColumnType, "Print", 50, True)
        self.tree.add_column(mforms.StringColumnType, "Diagram", 300, False)
        self.tree.end_columns()
        box.add(self.tree, True, True)

        for d in model.diagrams:
            node = self.tree.add_node()
            node.set_bool(0, True)
            node.set_string(1, d.name)

        help_text = """The following variables will be substituted:
$document, $doc_version, $doc_author, $doc_project, $doc_date_changed, $doc_date_created, $diagram, $timestamp, $page, $total_pages, $doc_page, $doc_total_pages
            """
        hbox = mforms.newBox(True)
        hbox.set_spacing(8)
        label = mforms.newLabel("Header Text:", True)
        label.set_size(100, -1)
        hbox.add(label, False, True)
        self.header = mforms.newTextEntry()
        self.header.set_tooltip(help_text)
        hbox.add(self.header, True, True)
        box.add(hbox, False, True)
        self.header.set_value(self.model.options.get("wb.PrintModel:HeaderText", "$timestamp, $document - $diagram (part $page of $total_pages)"))

        hbox = mforms.newBox(True)
        hbox.set_spacing(8)
        label = mforms.newLabel("Footer Text:", True)
        label.set_size(100, -1)
        hbox.add(label, False, True)
        self.footer = mforms.newTextEntry()
        self.footer.set_tooltip(help_text)
        hbox.add(self.footer, True, True)
        box.add(hbox, False, True)
        self.footer.set_value(self.model.options.get("wb.PrintModel:FooterText", "$doc_page of $doc_total_pages"))

        hbox = mforms.newBox(True)
        hbox.set_spacing(8)
        label = mforms.newLabel("File:", True)
        label.set_size(100, -1)
        hbox.add(label, False, True)
        self.file = mforms.newTextEntry()
        hbox.add(self.file, True, True)
        self.browse = mforms.newButton()
        self.browse.set_text("...")
        self.browse.add_clicked_callback(self.do_browse)
        self.browse.enable_internal_padding(False)
        hbox.add(self.browse, False, True)
        box.add(hbox, False, True)

        self.file.set_value(self.model.options.get("wb.PrintModel:Path", os.path.join(mforms.Utilities.get_special_folder(mforms.Desktop), "model.pdf")))

        hbox = mforms.newBox(True)
        hbox.set_spacing(8)
        label = mforms.newLabel("Format:", True)
        label.set_size(100, -1)
        hbox.add(label, False, True)
        self.format = mforms.newSelector()
        self.format.add_item("PDF")
        self.format.add_item("PostScript File")
        self.format.add_changed_callback(self.format_changed)
        hbox.add(self.format, True, True)
        box.add(hbox, False, True)

        hbox = mforms.newBox(True)
        hbox.set_spacing(8)
        self.ok = mforms.newButton()
        self.ok.set_text("OK")
        self.cancel = mforms.newButton()
        self.cancel.set_text("Cancel")
        mforms.Utilities.add_end_ok_cancel_buttons(hbox, self.ok, self.cancel)
        box.add_end(hbox, False, True)

        self.set_content(box)
        self.set_size(500, 400)


    def do_browse(self):
        fc = mforms.FileChooser(mforms.Form.main_form(), mforms.SaveFile)
        fc.set_path(self.file.get_string_value())
        fc.set_title("Print to File...")
        fc.set_extensions("PDF files (*.pdf)|*.pdf|PostScript Files (*.ps)|*.ps", ".pdf")
        if fc.run_modal():
            self.file.set_value(fc.get_path())


    def format_changed(self):
        path = self.file.get_string_value()
        format = [".pdf", ".ps"][self.format.get_selected_index()]
        if not path.endswith(format):
            path = os.path.splitext(path)[0]+format
            self.file.set_value(path)


    def run(self):
        if self.run_modal(self.ok, self.cancel):
            l = grt.List("object", "model.Diagram")
            for i, d in enumerate(self.model.diagrams):
                if self.tree.node_at_row(i).get_bool(0):
                    l.append(d)

            if not l:
                mforms.App.get().set_status_text("No diagrams selected to print")
                return

            file = self.file.get_string_value()
            format = [".pdf", ".ps"][self.format.get_selected_index()]
            header = self.header.get_string_value()
            footer = self.footer.get_string_value()

            self.model.options["wb.PrintModel:Path"] = file
            self.model.options["wb.PrintModel:HeaderText"] = header
            self.model.options["wb.PrintModel:FooterText"] = footer


            if not file:
                mforms.App.get().set_status_text("Invalid path")

            if not file.endswith(format):
                file = file + format

            mforms.App.get().set_status_text("Exporting diagrams to %s..." % file)

            def replace_variables(text, model):
                text = text.replace("$document", model.owner.info.caption)
                text = text.replace("$doc_version", model.owner.info.version)
                text = text.replace("$doc_author", model.owner.info.author)
                text = text.replace("$doc_project", model.owner.info.project)
                text = text.replace("$doc_date_changed", model.owner.info.dateChanged)
                text = text.replace("$doc_date_created", model.owner.info.dateCreated)
                import time
                text = text.replace("$timestamp", time.ctime())
                return text

            options = {
                "header_text" : replace_variables(header, self.model),
                "footer_text" : replace_variables(footer, self.model),
            }
            grt.modules.WbPrinting.printDiagramsToFile(l, file, format[1:], options)

            mforms.App.get().set_status_text("Exported %i diagrams to %s." % (len(l), file))



@ModuleInfo.plugin("wb.model.print_model", caption= "Print Model to File", input= [wbinputs.currentModel()])
@ModuleInfo.export(grt.INT, grt.classes.model_Model)
def printModel(model):
    dlg = PrintToFileDialog(model)
    dlg.run()
    return 0


# Quick impl for Rename Diagram... item in modeling overview
@ModuleInfo.plugin("wb.model.rename_diagram", caption= "Rename Diagram...", input= [wbinputs.selectedDiagram()], pluginMenu="Overview")
@ModuleInfo.export(grt.INT, grt.classes.model_Diagram)
def renameDiagram(diagram):
    ret, name = mforms.Utilities.request_input("Rename Diagram", "Enter new name for the diagram", diagram.name)
    if ret:
        grt.modules.Workbench.startTrackingUndo()
        diagram.name = name
        grt.modules.Workbench.finishTrackingUndo("Rename diagram to %s" % name)
    return 0



# Copyright (c) 2012, 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import random
import os.path

from wb import DefineModule
from workbench import db_driver
from workbench.utils import find_object_with_name
from workbench.exceptions import NotConnectedError

from db_generic_re_grt import GenericReverseEngineering

import SQLEXT as constant

import grt


ModuleInfo = DefineModule(name= "DbMsAccessRE", author= "Oracle Corp.", version="1.0")


def normalize_schema_name(path):
    """Turns an Access DB file path into something resembling a schema name"""
    return os.path.splitext(os.path.basename(path))[0]


class MsAccessReverseEngineering(GenericReverseEngineering):
    _connections = {}


    @classmethod
    def check_interruption(cls):
        if grt.query_status():
            raise grt.UserInterrupt()


    @classmethod
    def find_datatype_object(cls, catalog, datatype_name):
        ''' Finds the datatype object corresponding to the given datatype name.

        Returns: a tuple of the form (is_simple_datatype, datatype) where:
            is_simple_datatype: True if the datatype was found among the simple datatypes for
                                its corresponding RDBMS
            datatype:           The actual datatype object. None if not found
        '''
        simple_types = cls._rdbms.simpleDatatypes
        user_types = catalog.userDatatypes
        for simple_type in simple_types:
            if datatype_name == simple_type.name or datatype_name in simple_type.synonyms:
                return (True, simple_type)
        for user_type in user_types:
            if datatype_name == user_type.name:
                return (False, user_type)
        return (False, None)

    @classmethod
    def get_connection(cls, connection_object):
        if connection_object.__id__ in cls._connections:
            return cls._connections[connection_object.__id__]["connection"]
        else:
            raise NotConnectedError("No open connection to %s" % connection_object.hostIdentifier)

    # Note: try to avoid executing SQL code within this module
    @classmethod
    def execute_query(cls, connection_object, query, *args, **kwargs):
        """Retrieves a connection and executes the given query returning a cursor to iterate over results.

        The remaining positional and keyword arguments are passed with the query to the execute function
        """
        return cls.get_connection(connection_object).cursor().execute(query, *args, **kwargs)


    @classmethod
    def initializeDBMSInfo(cls, xml_data_path):
        cls._rdbms = grt.unserialize(os.path.join(ModuleInfo.moduleDataDirectory, xml_data_path))
        grt.root.wb.rdbmsMgmt.rdbms.append(cls._rdbms)
        return cls._rdbms

    
    @classmethod
    def getDataSourceNames(cls):
        result = grt.List(grt.STRING)
        import pyodbc
        sources = pyodbc.dataSources()
        for key, value in sources.items():
            result.append("%s|%s (%s)" % (key, key, value))
        return result


    @classmethod
    def getSupportedObjectTypes(cls):
        return [("tables", "db.Table", "Tables"), 
               ]


    @classmethod
    def quoteIdentifier(cls, name):
        return '"%s"' % name.replace('"', '\"')

    @classmethod
    def fullyQualifiedObjectName(cls, obj):
        owner = obj.owner
        if owner and isinstance(owner, grt.classes.db_Schema):
            return cls.quoteIdentifier(owner.name)+"."+cls.quoteIdentifier(obj.name)
        return cls.quoteIdentifier(obj.name)


    #########  Connection related functions #########
    @classmethod
    def connect(cls, connection, password):
        '''Establishes a connection to the server and stores the connection object in the connections pool.

        It first looks for a connection with the given connection parameters in the connections pool to
        reuse existent connections. If such connection is found it queries the server to ensure that the
        connection is alive and reestablishes it if is dead. If no suitable connection is found in the
        connections pool, a new one is created and stored in the pool.

        Parameters:
        ===========
            connection:  an object of the class db_mgmt_Connection storing the parameters
                         for the connection.
            password:    a string with the password to use for the connection.
        '''
        try:
            con = cls.get_connection(connection)
            try:
                if not con.cursor().execute('SELECT 1'):
                    raise Exception("connection error")
            except Exception, exc:
                grt.send_info("Connection to %s apparently lost, reconnecting..." % connection.hostIdentifier)
                raise NotConnectedError("Connection error")
        except NotConnectedError, exc:
            grt.send_info("Connecting to %s..." % connection.hostIdentifier)
            con = db_driver.connect(connection, password)
            if not con:
                grt.send_error('Connection failed', str(exc))
                raise
            grt.send_info("Connected")
            cls._connections[connection.__id__] = {"connection": con}
        return 1


    @classmethod
    def disconnect(cls, connection):
        if connection.__id__ in cls._connections:
            cls._connections[connection.__id__]['connection'].close()
            del cls._connections[connection.__id__]
        return 0

    @classmethod
    def isConnected(cls, connection):
        return 1 if connection.__id__ in cls._connections else 0


    #########  Exploratory functions (these only return useful info without reverse engineering) #########

        
    @classmethod
    def getDriverDBMSName(cls, connection):
        if connection.driver.driverLibraryName != 'pyodbc':
            return ''
        import pyodbc
        return cls.get_connection(connection).getinfo(pyodbc.SQL_DBMS_NAME)
    
    @classmethod
    def getTargetDBMSName(cls):
        return 'MsAccess'

    @classmethod
    def getServerVersion(cls, connection):
        """Returns a GrtVersion instance containing information about the server version."""
        
        # Note: Not implemented. This returns a predefined default server version for compatibility sake.
        version = grt.classes.GrtVersion()
        version.majorNumber, version.minorNumber, version.releaseNumber, version.buildNumber = 1, 0, 0, 0
        return version

    @classmethod
    def getCatalogNames(cls, connection):
        return ["def"]

    @classmethod
    def getSchemaNames(cls, connection, catalog_name):
        """Returns a list of the available schemas.

        [NOTE] This will in fact return the name of the database file we are connected to.
        """
        res = list(set(normalize_schema_name(row[0]) for row in cls.get_connection(connection).cursor().tables())) 
        assert len(res) == 1
        return res


    @classmethod
    def getTableNames(cls, connection, catalog_name, schema_name):
        return list(set(row.table_name for row in 
                    cls.get_connection(connection).cursor().tables() if row.table_type=='TABLE'))


    @classmethod
    def getViewNames(cls, connection, catalog_name, schema_name):
        return list(set(row.table_name for row in 
                    cls.get_connection(connection).cursor().tables() if row.table_type=='VIEW'))


    @classmethod
    def getTriggerNames(cls, connection, catalog_name, schema_name):
        return []


    @classmethod
    def getProcedureNames(cls, connection, catalog_name, schema_name):
        return list(set(row.procedure_name for row in 
                    cls.get_connection(connection).cursor().procedures() ))


    @classmethod
    def getFunctionNames(cls, connection, catalog_name, schema_name):
        return []


    #########  Reverse Engineering functions #########

    @classmethod
    def reverseEngineer(cls, connection, catalog_name, schemata_list, context):
        grt.send_progress(0, "Reverse engineering catalog information")
        cls.check_interruption()
        catalog = cls.reverseEngineerCatalog(connection, catalog_name)

        # calculate total workload 1st
        grt.send_progress(0.1, 'Preparing...')
        table_count_per_schema = {}
        view_count_per_schema = {}
        routine_count_per_schema = {}
        trigger_count_per_schema = {}
        total_count_per_schema = {}

        get_tables = context.get("reverseEngineerTables", True)
        get_triggers = context.get("reverseEngineerTriggers", True)
        get_views = context.get("reverseEngineerViews", True)
        get_routines = context.get("reverseEngineerRoutines", True)

        # 10% of the progress is for preparation
        total = 1e-10  # total should not be zero to avoid DivisionByZero exceptions
        i = 0.0
        accumulated_progress = 0.1
        for schema_name in schemata_list:
            cls.check_interruption()
            table_count_per_schema[schema_name] = len(cls.getTableNames(connection, catalog_name, schema_name)) if get_tables else 0
            view_count_per_schema[schema_name] = len(cls.getViewNames(connection, catalog_name, schema_name)) if get_views else 0
            cls.check_interruption()
            routine_count_per_schema[schema_name] = len(cls.getProcedureNames(connection, catalog_name, schema_name)) + len(cls.getFunctionNames(connection, catalog_name, schema_name)) if get_routines else 0
            trigger_count_per_schema[schema_name] = len(cls.getTriggerNames(connection, catalog_name, schema_name)) if get_triggers else 0

            total_count_per_schema[schema_name] = (table_count_per_schema[schema_name] + view_count_per_schema[schema_name] +
                                                   routine_count_per_schema[schema_name] + trigger_count_per_schema[schema_name] + 1e-10)
            total += total_count_per_schema[schema_name]

            grt.send_progress(accumulated_progress + 0.1 * (i / (len(schemata_list) + 1e-10) ), "Gathered stats for %s" % schema_name)
            i += 1.0

        # Now take 60% in the first pass of reverse engineering:
        accumulated_progress = 0.2
        for schema_name in schemata_list:
            schema_progress_share = 0.6 * (total_count_per_schema.get(schema_name, 0.0) / total)
            schema = find_object_with_name(catalog.schemata, schema_name) 

            if schema:
                # Reverse engineer tables:
                step_progress_share = schema_progress_share * (table_count_per_schema[schema_name] / (total_count_per_schema[schema_name] + 1e-10))
                if get_tables:
                    cls.check_interruption()
                    grt.send_info('Reverse engineering tables from %s' % schema_name)
                    grt.begin_progress_step(accumulated_progress, accumulated_progress + step_progress_share)
                    # Remove previous first pass marks that may exist if the user goes back and attempt rev eng again:
                    progress_flags = cls._connections[connection.__id__].setdefault('_rev_eng_progress_flags', set())
                    progress_flags.discard('%s_tables_first_pass' % schema_name)
                    cls.reverseEngineerTables(connection, schema)
                    grt.end_progress_step()
        
                accumulated_progress += step_progress_share
                grt.send_progress(accumulated_progress, 'First pass of table reverse engineering for schema %s completed!' % schema_name)
        
                # Reverse engineer views:
                step_progress_share = schema_progress_share * (view_count_per_schema[schema_name] / (total_count_per_schema[schema_name] + 1e-10))
                if get_views:
                    cls.check_interruption()
                    grt.send_info('Reverse engineering views from %s' % schema_name)
                    grt.begin_progress_step(accumulated_progress, accumulated_progress + step_progress_share)
                    cls.reverseEngineerViews(connection, schema)
                    grt.end_progress_step()
        
                accumulated_progress += step_progress_share
                grt.send_progress(accumulated_progress, 'Reverse engineering of views for schema %s completed!' % schema_name)
        
                # Reverse engineer routines:
                step_progress_share = schema_progress_share * (routine_count_per_schema[schema_name] / (total_count_per_schema[schema_name] + 1e-10))
                if get_routines:
                    cls.check_interruption()
                    grt.send_info('Reverse engineering routines from %s' % schema_name)
                    grt.begin_progress_step(accumulated_progress, accumulated_progress + step_progress_share)
                    grt.begin_progress_step(0.0, 0.5)
                    cls.reverseEngineerProcedures(connection, schema)
                    cls.check_interruption()
                    grt.end_progress_step()
                    grt.begin_progress_step(0.5, 1.0)
                    reverseEngineerFunctions(connection, schema)
                    grt.end_progress_step()
                    grt.end_progress_step()
        
                accumulated_progress += step_progress_share
                grt.send_progress(accumulated_progress, 'Reverse engineering of routines for schema %s completed!' % schema_name)
        
                # Reverse engineer triggers:
                step_progress_share = schema_progress_share * (trigger_count_per_schema[schema_name] / (total_count_per_schema[schema_name] + 1e-10))
                if get_triggers:
                    cls.check_interruption()
                    grt.send_info('Reverse engineering triggers from %s' % schema_name)
                    grt.begin_progress_step(accumulated_progress, accumulated_progress + step_progress_share)
                    cls.reverseEngineerTriggers(connection, schema)
                    grt.end_progress_step()
        
                accumulated_progress = 0.8
                grt.send_progress(accumulated_progress, 'Reverse engineering of triggers for schema %s completed!' % schema_name)
            else:  # No schema with the given name was found
                grt.send_warning('The schema %s was not found in the catalog %s. Skipping it.' % (schema_name, catalog_name) )
                
        # Now the second pass for reverse engineering tables:
        if get_tables:
            total_tables = sum(table_count_per_schema[schema.name] for schema in catalog.schemata if schema.name in schemata_list)
            for schema in catalog.schemata:
                if schema.name not in schemata_list:
                    continue
                cls.check_interruption()
                step_progress_share = 0.2 * (table_count_per_schema[schema.name] / (total_tables + 1e-10))
                grt.send_info('Reverse engineering foreign keys for tables in schema %s' % schema.name)
                grt.begin_progress_step(accumulated_progress, accumulated_progress + step_progress_share)
                cls.reverseEngineerTables(connection, schema)
                grt.end_progress_step()
        
                accumulated_progress += step_progress_share
                grt.send_progress(accumulated_progress, 'Second pass of table reverse engineering for schema %s completed!' % schema_name)
            

        grt.send_progress(1.0, 'Reverse engineering completed!')
        return catalog


    @classmethod
    def reverseEngineerUserDatatypes(cls, connection, catalog):
        catalog.simpleDatatypes.remove_all()
        for type_row in cls.get_connection(connection).cursor().getTypeInfo(catalog=catalog.name):  # FIXME: there are duplicated names in this resultset
            simple_datatype = grt.classes.db_SimpleDatatype()
            simple_datatype.name = type_row[0]
            simple_datatype.characterMaximumLength = simple_datatype.characterOctetLength = simple_datatype.numericPrecision = simple_datatype.dateTimePrecision = type_row[2] if isinstance(type_row[2], int) else -1
            if isinstance(type_row[17], int):
                simple_datatype.numericPrecisionRadix = type_row[17]
            if isinstance(type_row[14], int):
                simple_datatype.numericScale = type_row[14]
            parameter_format_type_mapping = { 0: 0, # none
                                              1: 2, # [(n)]
                                              2: 6, # [(m[, n])]
                                            }
            if type_row[5] is not None:  # parameter format
                simple_datatype.parameterFormatType = parameter_format_type_mapping.get(len(type_row[5].split(',')), 0)
            else:
                simple_datatype.parameterFormatType = 0

            simple_datatype.needsQuotes = type_row[3] in ["N'", "'"]
            
            catalog.simpleDatatypes.append(simple_datatype)


    @classmethod
    def reverseEngineerCatalog(cls, connection, catalog_name):
        catalog = grt.classes.db_Catalog()
        catalog.name = catalog_name
        
        cls.reverseEngineerUserDatatypes(connection, catalog)

        schemata_names = cls.getSchemaNames(connection, catalog_name) or ['']
        catalog.schemata.remove_all()
        for schema_name in schemata_names:
            schema = grt.classes.db_Schema()
            schema.name = schema_name
            schema.owner = catalog
            if hasattr(cls, 'reverseEngineerSequences'):
                cls.reverseEngineerSequences(connection, schema)
            catalog.schemata.append(schema)
        return catalog

    @classmethod
    def reverseEngineerTables(cls, connection, schema):
        # Since there are some reverse engineering stages that requires all table names and table columns
        # in the database to be set, these should be done after a first pass that rev engs their requirements
        progress_flags = cls._connections[connection.__id__].setdefault('_rev_eng_progress_flags', set())
        is_first_pass = not ('%s_tables_first_pass' % schema.name) in progress_flags

        if is_first_pass:
            catalog = schema.owner
            schema.tables.remove_all()
            table_names = cls.getTableNames(connection, catalog.name, schema.name)
            getCommentForTable = cls.getCommentForTable if hasattr(cls, 'getCommentForTable') else lambda conn, tbl:''
            total = len(table_names) + 1e-10
            i = 0.0
            for table_name in table_names:
                grt.send_progress(i / total, 'Retrieving table %s.%s...' % (schema.name, table_name))
                table = grt.classes.db_Table()
                table.name = table_name
                schema.tables.append(table)
                table.owner = schema
                table.comment = getCommentForTable(connection, table)
        
                cls.reverseEngineerTableColumns(connection, table)
                cls.reverseEngineerTablePK(connection, table)
                cls.reverseEngineerTableIndices(connection, table)
        
                i += 1.0
            progress_flags.add('%s_tables_first_pass' % schema.name)
        else:  # Second pass
            i = 0.0
            total = len(schema.tables) + 1e-10
            cls._connections[connection.__id__]['fk_names'] = {}
            for table in schema.tables:
                cls.reverseEngineerTableFKs(connection, table)
                grt.send_progress(i / total, 'Reverse engineering of foreign keys in table %s.%s completed' % (schema.name, table.name))
                i += 1.0

        return 0

    @classmethod
    def reverseEngineerTableColumns(cls, connection, table):
        schema = table.owner
        catalog = schema.owner

        simple_datatypes_list = [ datatype.name.upper() for datatype in catalog.simpleDatatypes ]
        user_datatypes_list   = [ datatype.name.upper() for datatype in catalog.userDatatypes ]

        odbc_datatypes = dict( (dtype.data_type, dtype.type_name) for dtype in cls.get_connection(connection).cursor().getTypeInfo() )

        table_columns = cls.get_connection(connection).cursor().columns(table=table.name)
        for column_info in table_columns:
            column = grt.classes.db_Column()
            column.name = column_info[3]  # column_name
            column.isNotNull = column_info[17] != 'YES'  # is_nullable
            column.length = column_info[6]  # column_size
            column.scale = column_info[8]  # decimal_digits
            column.precision = column_info[6]  # column_size

            datatype = None
            try:
                type_name = odbc_datatypes[column_info[4]].upper()  # data_type
                datatype = simple_datatypes_list.index(type_name)
            except (KeyError, ValueError):
                try:
                    user_datatype = catalog.userDatatypes[user_datatypes_list.index(type_name)]
                except (ValueError, TypeError, NameError):
                    user_datatype = None
                    datatype = simple_datatypes_list.index('VARCHAR')
                    column.length = 255
                    msg = 'Column datatype "%s" for column "%s" in table "%s.%s" reverse engineered as VARCHAR(255)' % (type_name, column.name, schema.name, table.name)
                    grt.send_warning('%s reverseEngineerTableColumns: ' % cls.getTargetDBMSName() + msg)
                else:
                    datatype = None
                    column.userType = user_datatype

            if isinstance(datatype, int):
                column.simpleType = catalog.simpleDatatypes[datatype]

            table.addColumn(column)

        return 0


    @classmethod
    def reverseEngineerTablePK(cls, connection, table):
        """Reverse engineers the primary key for the given table."""

        if len(table.columns) == 0:  # Table must have columns reverse engineered before we can rev eng its primary key
            grt.send_error('Migration: reverseEngineerTablePKAndIndices: Reverse engineer of table %s was attempted but the table has no columns attribute' % table.name)
            return 1
        
        pk_index_name = 'PrimaryKey'

        indices_dict = {}  # Map the indices names to their respective columns:
        for row in cls.get_connection(connection).cursor().statistics(table=table.name):
            if row.type == constant.SQL_TABLE_STAT:  # this entry is not an index
                continue
            indices_dict.setdefault(row.index_name, []).append(row)

        for index_name, row_list in indices_dict.iteritems():
            index = grt.classes.db_Index()
            index.name = index_name
            index.isPrimary = 1 if index_name == pk_index_name else 0
            index.unique = not row_list[0].non_unique
            index.indexType = 'UNIQUE' if index.unique else 'INDEX'
    #        index.hasFilter = False  # TODO: Find out if there's a way to determine this

            skip = False
            # Get the columns for the index:
            for row in sorted(row_list, key=lambda elem: elem[7]):  # Sorted by ordinal_position
                column = find_object_with_name(table.columns, row.column_name)
                if column:
                    # skip indexes on LONGCHAR columns
                    if column.simpleType.name in ["LONGCHAR"]:
                        grt.send_warning("Migration: reverseEngineerTable: Skipping index %s.%s on a %s column\n" % (table.name, column.name, column.simpleType.name)) 
                        skip = True
                    else:
                        index_column = grt.classes.db_IndexColumn()
                        index_column.name = index_name + '.' + row.column_name
                        index_column.referencedColumn = column
                        index.columns.append(index_column)
                        if not column.isNotNull and index.isPrimary:
                            column.isNotNull = 1
                            grt.send_warning("Migration: reverseEngineerTablePK: column %s.%s was changed to NOT NULL because it's a Primary Key column\n" % (column.owner.name, column.name))
                else:
                    grt.send_warning("Migration: reverseEngineerTablePK: could not find column %s, belonging to key %s. Key will be skipped\n" % (row.column_name, index_name))
                    skip = True
            if not skip:
                table.addIndex(index)
                if index.isPrimary:
                    table.primaryKey = index

        return 0


    @classmethod
    def reverseEngineerTableIndices(cls, connection, table):
        pass  # Indices already reverse engineered in reverseEngineerTablePK


    @classmethod
    def reverseEngineerTableFKs(cls, connection, table):
        """Reverse engineers the foreign keys for the given table."""
        
        def get_update_action(grbit):
            if grbit & 256:
                return "CASCADE"
            else:
                return "RESTRICT"

        def get_delete_action(grbit):
            if grbit & 4352:
                return "CASCADE"
            else:
                return "RESTRICT"
            
        def process_fk(catalog, table, fk_name, fk_rows):
            foreign_key = grt.classes.db_ForeignKey()
            if fk_name in cls._connections[connection.__id__]['fk_names']:
                while True:
                    suffix = '_%06d' % random.randint(0, 999999)
                    if fk_name + suffix not in cls._connections[connection.__id__]['fk_names']:
                        break
                fk_name += suffix
            foreign_key.name = fk_name
            foreign_key.owner = table
            foreign_key.deleteRule = get_delete_action(fk_rows[0].grbit)
            foreign_key.updateRule = get_update_action(fk_rows[0].grbit)
            foreign_key.modelOnly = 0
            
            # Find the referenced table:
            foreign_key.referencedTable = find_object_with_name(catalog.schemata[0].tables, fk_rows[0].szReferencedObject)
            if not foreign_key.referencedTable:
                grt.send_error('Migration: reverseEngineerTableFKs: Table "%s" not found in schemata "%s"' % (fk_rows[0].szReferencedObject, catalog.schemata[0].name) )
                return 1
            
            for fk_row in fk_rows:
                column = find_object_with_name(table.columns, fk_row.szColumn)
                if not column:
                    grt.send_error('Migration: reverseEngineerTableFKs: Column "%s" not found in table "%s"' % (fk_row.szColumn, table.name) )
                    continue

                ref_column = find_object_with_name(foreign_key.referencedTable.columns, fk_row.szReferencedColumn)
                if not ref_column:
                    grt.send_error('Migration: reverseEngineerTableFKs: Column "%s" not found in table "%s"' % (fk_row.szReferencedColumn, foreign_key.referencedTable.name) )
                    continue
                
                foreign_key.columns.append(column)
                foreign_key.referencedColumns.append(ref_column)

            # Find and delete indexes that are identical to FKs
            for index in reversed(table.indices):
                if table.primaryKey != index and len(index.columns) == len(foreign_key.columns):
                    match = True
                    for i, col in enumerate(index.columns):
                        if foreign_key.columns[i] != col.referencedColumn:
                            match = False
                            break
                    if match:
                        grt.send_warning("Migration: reverseEngineerTable: Skipping duplicate index %s from table %s\n" % (col.name, table.name))
                        table.indices.remove(index)

            cls._connections[connection.__id__]['fk_names'][foreign_key.name] = table 
            table.foreignKeys.append(foreign_key)
                

        if len(table.columns) == 0:
            grt.send_error('Migration: reverseEngineerTableFKs: Reverse engineer of table %s was attempted but the table has no columns attribute' % table.name)
            return 1    # Table must have columns reverse engineered before we can rev eng its indices

        catalog = table.owner.owner
        table.foreignKeys.remove_all()
        fk_dict = {}  # Map the foreign key names to their respective columns:

        import pyodbc
        try:
            for row in cls.get_connection(connection).cursor().execute("SELECT * FROM MSysRelationships WHERE szObject = ?", (table.name,)):
                fk_dict.setdefault(row.szRelationship, []).append(row)
        except pyodbc.ProgrammingError, e:
            if e.args[0] == '42000':
                grt.send_error("\n\nMigration: Could not read from System Tables. You must grant SELECT access on all system tables for the database.")
                return 1
            raise

        for fk_name, fk_columns in fk_dict.iteritems():
            process_fk(catalog, table, fk_name, fk_columns)
        return 0


    @classmethod
    def reverseEngineerViews(cls, connection, schema):
        for view_name in cls.getViewNames(connection, schema.owner.name, schema.name):
            grt.send_info('%s reverseEngineerViews: Cannot reverse engineer view "%s"' % (cls.getTargetDBMSName(), view_name))
        return 0


    @classmethod
    def reverseEngineerProcedures(cls, connection, schema):
        # Unfortunately it seems that there's no way to get the SQL definition of a store procedure/function with ODBC
        for procedure_name in cls.getProcedureNames(connection, schema.owner.name, schema.name):
            grt.send_info('%s reverseEngineerProcedures: Cannot reverse engineer procedure "%s"' % (cls.getTargetDBMSName(), procedure_name))
        return 0


    @classmethod
    def reverseEngineerFunctions(cls, connection, schema):
        # Unfortunately it seems that there's no way to get the SQL definition of a store procedure/function with ODBC
        for function_name in cls.getFunctionNames(connection, schema.owner.name, schema.name):
            grt.send_info('%s reverseEngineerFunctions: Cannot reverse engineer function "%s"' % (cls.getTargetDBMSName(), function_name))
        return 0


    @classmethod
    def reverseEngineerTriggers(cls, connection, schema):
        # Unfortunately it seems that there's no way to get the SQL definition of a trigger with ODBC
        for trigger_name in cls.getTriggerNames(connection, schema.owner.name, schema.name):
            grt.send_info('%s reverseEngineerTriggers: Cannot reverse engineer trigger "%s"' % (cls.getTargetDBMSName(), trigger_name))
        return 0


    @classmethod
    def resetProgressFlags(cls, connection):
        cls._connections[connection.__id__]['_rev_eng_progress_flags'] = set()
        return 0
###############################################################################################################

@ModuleInfo.export(grt.classes.db_mgmt_Rdbms)
def initializeDBMSInfo():
    return MsAccessReverseEngineering.initializeDBMSInfo('msaccess_rdbms_info.xml')

@ModuleInfo.export((grt.LIST, grt.STRING))
def getDataSourceNames():
    return MsAccessReverseEngineering.getDataSourceNames()


@ModuleInfo.export(grt.LIST)
def getSupportedObjectTypes():
    return MsAccessReverseEngineering.getSupportedObjectTypes()


@ModuleInfo.export(grt.STRING, grt.STRING)
def quoteIdentifier(name):
    return MsAccessReverseEngineering.quoteIdentifier(name)


@ModuleInfo.export(grt.STRING, grt.classes.GrtNamedObject)
def fullyQualifiedObjectName(obj):
    return MsAccessReverseEngineering.fullyQualifiedObjectName(obj)


@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.STRING)
def connect(connection, password):
    return MsAccessReverseEngineering.connect(connection, password)


@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection)
def disconnect(connection):
    return MsAccessReverseEngineering.disconnect(connection)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection)
def isConnected(connection):
    return MsAccessReverseEngineering.isConnected(connection)


@ModuleInfo.export(grt.STRING, grt.classes.db_mgmt_Connection)
def getDriverDBMSName(connection):
    return MsAccessReverseEngineering.getDriverDBMSName(connection)
    
@ModuleInfo.export(grt.STRING)
def getTargetDBMSName():
    return MsAccessReverseEngineering.getTargetDBMSName()

@ModuleInfo.export(grt.classes.GrtVersion, grt.classes.db_mgmt_Connection)
def getServerVersion(connection):
    return MsAccessReverseEngineering.getServerVersion(connection)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection)
def getCatalogNames(connection):
    return MsAccessReverseEngineering.getCatalogNames(connection)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING)
def getSchemaNames(connection, catalog_name):
    return MsAccessReverseEngineering.getSchemaNames(connection, catalog_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getTableNames(connection, catalog_name, schema_name):
    return MsAccessReverseEngineering.getTableNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getViewNames(connection, catalog_name, schema_name):
    return MsAccessReverseEngineering.getViewNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getTriggerNames(connection, catalog_name, schema_name):
    return MsAccessReverseEngineering.getTriggerNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getProcedureNames(connection, catalog_name, schema_name):
    return MsAccessReverseEngineering.getProcedureNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.LIST, grt.classes.db_mgmt_Connection, grt.STRING, grt.STRING)
def getFunctionNames(connection, catalog_name, schema_name):
    return MsAccessReverseEngineering.getFunctionNames(connection, catalog_name, schema_name)

@ModuleInfo.export(grt.classes.db_Catalog, grt.classes.db_mgmt_Connection, grt.STRING, (grt.LIST, grt.STRING), grt.DICT)
def reverseEngineer(connection, catalog_name, schemata_list, context):
    return MsAccessReverseEngineering.reverseEngineer(connection, catalog_name, schemata_list, context)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Catalog)
def reverseEngineerUserDatatypes(connection, catalog):
    return MsAccessReverseEngineering.reverseEngineerUserDatatypes(connection, catalog)

@ModuleInfo.export(grt.classes.db_Catalog, grt.classes.db_mgmt_Connection, grt.STRING)
def reverseEngineerCatalog(connection, catalog_name):
    return MsAccessReverseEngineering.reverseEngineerCatalog(connection, catalog_name)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerTables(connection, schema):
    return MsAccessReverseEngineering.reverseEngineerTables(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerViews(connection, schema):
    return MsAccessReverseEngineering.reverseEngineerViews(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerProcedures(connection, schema):
    return MsAccessReverseEngineering.reverseEngineerProcedures(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerFunctions(connection, schema):
    return MsAccessReverseEngineering.reverseEngineerFunctions(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection, grt.classes.db_Schema)
def reverseEngineerTriggers(connection, schema):
    return MsAccessReverseEngineering.reverseEngineerTriggers(connection, schema)

@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_Connection)
def resetProgressFlags(connection):
    return MsAccessReverseEngineering.resetProgressFlags(connection)




import option
import wb_admin_config_file_be
from wb_common import Users
#===============================================================================
#
#===============================================================================
class DummyPwdHandler:
    def get_password_for(self, name):
        return ""

    def reset_password_for(self, name):
        pass

#===============================================================================
#
#===============================================================================
class HelperDummy:
    content = None
    def __init__(self, content):
        content = self.content

    def get_file_content(file_name, as_user=Users.CURRENT, user_password=None, skip_lines=0):
        return self.content

#===============================================================================
#
#===============================================================================
class CtrlBEDummy:
    def __init__(self):
        self.helper = HelperDummy([])
        self.password_handler = DummyPwdHandler()

#===============================================================================
#
#===============================================================================
class DummyProfile:
    @property
    def target_is_windows(self):
        return False

    @property
    def config_file_section(self):
        return "mysqld"

    @property
    def admin_enabled(self):
        return True

    @property
    def config_file_section(self):
        return "mysqld"


profile = DummyProfile()
#===============================================================================
#
#===============================================================================
def unit_test_1():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["#Start\n", "[mysqld]\n", "ansi\n"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    for name, opt in cfg.options.iteritems():
        values  = opt.get_values()
        default = opt.default_value()
        if len(values) > 0:
            if default is not True:
                if name == 'ansi' and values[0].enabled == True:
                    ret = True

    return (ret, "Reading simple config")

#-------------------------------------------------------------------------------
def unit_test_2():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["#Start\n", "[mysqld]\n", "binlog-do-db=db1\n", "binlog-do-db=db2\n"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    cfg_opts = {}
    for name, opt in cfg.options.iteritems():
        values  = opt.get_values()
        default = opt.default_value()
        if len(values) == 2:
            cfg_opts[id(opt)] = opt

    ret = len(cfg_opts) == 1 and len(cfg_opts.values()[0].get_values()) == 2

    values = cfg_opts.values()[0].get_values()
    ret = ret and values[0].value == "db1" and values[1].value == "db2"

    return (ret, "Reading simple config with multiline option")

#-------------------------------------------------------------------------------
def unit_test_3():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["#Start\n", "[mysqld]\n", "ansi\n"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("chroot")
    values = opt.get_values()
    values.append(option.Value("/var/ww/mydir", True))
    opt.set_values(values)

    new_lines = cfg.generate_file_content()
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(new_lines)

    opt = cfg.get_option("chroot")
    ret = opt.get_values()[0].value == '"/var/ww/mydir"'

    return (ret, "Adding dirname/filename option")

#-------------------------------------------------------------------------------
def unit_test_4():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = []
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False
    new_lines = cfg.generate_file_content()
    ret = '[mysqld]' in new_lines

    return (ret, "Parsing empty config")

#-------------------------------------------------------------------------------
def unit_test_5():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["#Start\n", "[mysqld]\n"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("binlog-do-db")
    values = opt.get_values()
    values.append(option.Value("db1", True))
    values.append(option.Value("db2", True))
    opt.set_values(values)

    new_lines = cfg.generate_file_content()
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(new_lines)

    ret =  'binlog-do-db = db2' in new_lines and 'binlog-do-db = db1' in new_lines

    new_lines = cfg.generate_file_content()
    ret =  ret and 'binlog-do-db = db2' in new_lines and 'binlog-do-db = db1' in new_lines

    return (ret, "Adding multi-line option")

#-------------------------------------------------------------------------------
def unit_test_6():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["#Start", "[mysql]", "port=3306"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("binlog-do-db")
    values = opt.get_values()
    values.append(option.Value("db1", True))
    values.append(option.Value("db2", True))
    opt.set_values(values)

    new_lines = cfg.generate_file_content()
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(new_lines)

    ret =  'binlog-do-db = db2' in new_lines and 'binlog-do-db = db1' in new_lines

    new_lines = cfg.generate_file_content()
    ret =  ret and 'binlog-do-db = db2' in new_lines and 'binlog-do-db = db1' in new_lines

    return (ret, "Adding multi-line option to non-empty file with missing [mysqld]")

#-------------------------------------------------------------------------------
def unit_test_7():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["#Start", "[mysql]", "port=3306", "[mysqld]", 'binlog-do-db = db1', 'binlog-do-db = db2']
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("binlog-do-db")
    values = opt.get_values()
    values[1].value = "db3"

    new_lines = cfg.generate_file_content()
    ret = 'binlog-do-db = db3' in new_lines and 'binlog-do-db = db1' in new_lines

    return (ret, "Changing multi-line option")

#-------------------------------------------------------------------------------
def unit_test_8():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["#Start", "[mysql]", "port=3306", "[mysqld]", 'binlog-do-db = db1', 'binlog-do-db = db2']
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("binlog-do-db")
    values = opt.get_values()
    values[1].enabled = False

    new_lines = cfg.generate_file_content()
    ret = 'binlog-do-db = db2' not in new_lines and 'binlog-do-db = db1' in new_lines

    return (ret, "Deleting one value from a multi-line option")


#-------------------------------------------------------------------------------
def unit_test_9():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["[mysqld]","local-infile = 0",'general-log-file="sdf1"','#general_log_file="sdf2"',"replicate-ignore-db='data1'","replicate-ignore-db='data2'",
"replicate-ignore-db='data3'","port=3306","port=3307"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("replicate-ignore-db")
    values = opt.get_values()
    ret = len(values) == 3

    return (ret, "Testing cfg")

#-------------------------------------------------------------------------------
def unit_test_10():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["[mysqld]","local-infile = 0",'general-log-file="sdf1"','#general_log_file="sdf2"',"replicate-ignore-db='data1'","replicate-ignore-db='data2'",
"replicate-ignore-db='data3'","port=3306","port=3307"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("local-infile")
    (c,r,a) = opt.get_changeset()
    ret = len(c) == len(r) == len(a) == 0

    opt = cfg.get_option("tmpdir")
    values = opt.get_values()
    values.append(option.Value("", True))
    opt.set_values(values)

    #print cfg.generate_file_content()

    return (ret, "Testing cfg")

#-------------------------------------------------------------------------------
def unit_test_11():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["[mysqld]","local-infile = 0",'general-log-file="sdf1"','#general_log_file="sdf2"',"replicate-ignore-db='data1'","replicate-ignore-db='data2'",
"replicate-ignore-db='data3'","port=3306","port=3307"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("skip-networking")
    opt.set_values([option.Value(None, True)])
    opt.set_values([option.Value(None, False)])
    (c, r, a) = opt.get_changeset()

    cont = cfg.generate_file_content()

    return (ret, "Testing cfg")

#-------------------------------------------------------------------------------
def unit_test_12():
    ctrl_be = CtrlBEDummy()
    ctrl_be.helper.content = ["[mysqld]","local-infile = 0",'general-log-file="sdf1"','#general_log_file="sdf2"',"replicate-ignore-db='data1'","replicate-ignore-db='data2'",
"replicate-ignore-db='data3'","port=3306","port=3307","skip-networking"]
    cfg = wb_admin_config_file_be.ConfigFileSource((5,1,8), ctrl_be, profile)
    cfg.file_name = "dummy_name"
    cfg.parse_file_contents(ctrl_be.helper.content)

    ret = False

    opt = cfg.get_option("skip-networking")
    values = opt.get_values()
    opt.set_values([option.Value(None, False)])
    (c, r, a) = opt.get_changeset()

    cont = cfg.generate_file_content()
    print cont

    return (ret, "Testing cfg")











#-------------------------------------------------------------------------------
results = []
tests = sorted(filter(lambda x: "unit_test" in x, dir()))

tests_numbers = sorted(map(lambda x: int(x[10:]), tests))

for testnr in tests_numbers:
    test = "unit_test_%i" % testnr
    print "--------------------------------------------------------------------"
    print "-- %s starting" % test
    ret = eval(test + "()")
    print "-- %s %s:   %s\n" % (test, str(ret[0]), ret[1])
    results.append((ret[0], ret[1], test))



print "============================== SUMMARY ==================================="
for res in results:
    ch = "+" if res[0] else "-"
    print "%s %s - %s" % (ch, res[2], res[1])
